import asyncio
import inspect
import json
import logging
import string
import time
import traceback
from enum import Enum
from pathlib import Path
from threading import Thread
from typing import Optional, Union, Any
from socket import timeout as socket_timeout
from warnings import warn

from pydantic import BaseModel, ValidationError
from starlette.requests import Request
from starlette.websockets import WebSocket
from starlette_admin.exceptions import ActionFailed
from kombu import Connection, Exchange, Queue, Message, Producer

from wiederverwendbar.logger.context import LoggingContext
from wiederverwendbar.starlette_admin.action_log.settings import ActionLogAdminSettings

LOGGER = logging.getLogger(__name__)
KOMBU_EXCHANGE_NAME = "action_log"


class _SubLoggerCommand:
    def __init__(self,
                 logger, allowed_logger_cls: list[type],
                 command: str,
                 **values):
        if not any([isinstance(logger, cls) for cls in allowed_logger_cls]):
            raise ValueError(f"Logger must be an instance of: {', '.join([cls.__name__ for cls in allowed_logger_cls])}.")
        self.logger = logger
        if isinstance(self.logger, ActionLogger):
            self.logger.send_command(ActionLoggerCommand(sub_logger="", command=command, value=values))
        else:
            self.logger.handle(self.logger.makeRecord(logger.name,
                                                      logging.NOTSET,
                                                      "",
                                                      0,
                                                      "",
                                                      (),
                                                      None,
                                                      extra={"handling_command": {"command": command, "values": values}}))


class StartCommand(_SubLoggerCommand):
    def __init__(self,
                 logger: Union["ActionSubLogger", logging.Logger],
                 steps: Optional[int] = None):
        super().__init__(logger=logger,
                         allowed_logger_cls=[ActionSubLogger, logging.Logger],
                         command="start",
                         steps=steps)


class StepCommand(_SubLoggerCommand):
    def __init__(self,
                 logger: Union["ActionSubLogger", logging.Logger],
                 step: int,
                 steps: Optional[int] = None):
        super().__init__(logger=logger,
                         allowed_logger_cls=[ActionSubLogger, logging.Logger],
                         command="step",
                         step=step,
                         steps=steps)


class NextStepCommand(_SubLoggerCommand):
    def __init__(self,
                 logger: Union["ActionSubLogger", logging.Logger],
                 steps: int = 1):
        super().__init__(logger=logger,
                         allowed_logger_cls=[ActionSubLogger, logging.Logger],
                         command="next_step",
                         steps=steps)


class IncreaseStepsCommand(_SubLoggerCommand):
    def __init__(self,
                 logger: Union["ActionSubLogger", logging.Logger],
                 steps: int):
        super().__init__(logger=logger,
                         allowed_logger_cls=[ActionSubLogger, logging.Logger],
                         command="increase_steps",
                         steps=steps)


class FormCommand(_SubLoggerCommand):
    def __init__(self,
                 logger: Union["ActionLogger", "ActionSubLogger", logging.Logger],
                 form: str,
                 submit_btn_text: Optional[str] = None,
                 abort_btn_text: Optional[str] = None,
                 default_values: Union[None, bool, dict[str, Any]] = None):
        if submit_btn_text is None:
            submit_btn_text = "OK"
        try:
            self.get_logger(logger)
        except ValueError:
            if default_values is None:
                raise ValueError(f"No action logger found. Did you use the {ActionSubLoggerContext.__name__} context manager? If not, you have to provide default values.")
        self.default_values: Union[None, bool, dict[str, Any]] = default_values

        super().__init__(logger=logger,
                         allowed_logger_cls=[ActionLogger, ActionSubLogger, logging.Logger],
                         command="form",
                         form=form,
                         submit_btn_text=submit_btn_text,
                         abort_btn_text=abort_btn_text)

    def __call__(self, timeout: Optional[float] = None) -> Union[bool, dict[str, Any]]:
        if isinstance(self.logger, ActionLogger) or isinstance(self.logger, ActionSubLogger):
            return asyncio.run(self.logger.form_data(timeout=timeout))
        elif isinstance(self.logger, logging.Logger):
            if self.default_values is None:
                raise ValueError("No default values provided.")
            return self.default_values
        else:
            raise ValueError("Logger must be an instance of ActionLogger, ActionSubLogger or logging.Logger.")

    @classmethod
    def get_logger(cls, logger: Union["ActionLogger", "ActionSubLogger", logging.Logger]) -> Union["ActionLogger", "ActionSubLogger"]:
        if isinstance(logger, ActionLogger) or isinstance(logger, ActionSubLogger):
            return logger
        action_sub_loggers = ActionSubLoggerContext.get_from_stack(inspect.stack())
        action_sub_logger_context: Optional[ActionSubLoggerContext] = None
        for action_sub_logger_context in action_sub_loggers:
            if isinstance(action_sub_logger_context, ActionSubLoggerContext):
                break
        if action_sub_logger_context is None:
            raise ValueError(f"No action logger found. Did you use the {ActionSubLoggerContext.__name__} context manager? If not, you have to provide default values.")
        return action_sub_logger_context.context_logger


class ConfirmCommand(FormCommand):
    def __init__(self,
                 logger: Union["ActionLogger", "ActionSubLogger", logging.Logger],
                 text: str,
                 submit_btn_text: Optional[str] = None,
                 form: Optional[str] = None):
        if form is None:
            form = f"""<form>
            <div class="mt-3">
                <p>{text}</p>
            </div>
            </form>"""
        super().__init__(logger=logger,
                         form=form,
                         submit_btn_text=submit_btn_text,
                         default_values=True)

    def __call__(self, timeout: Optional[float] = None) -> None:
        super().__call__(timeout=timeout)


class DownloadCommand(ConfirmCommand):
    def __init__(self,
                 logger: Union["ActionLogger", "ActionSubLogger", logging.Logger],
                 file_path: Union[str, Path],
                 text: str,
                 icon: Optional[str] = None,
                 submit_btn_text: Optional[str] = None,
                 form: Optional[str] = None):
        # check if file exists
        if type(file_path) is str:
            file_path = Path(file_path)
        if not file_path.is_file():
            raise ValueError(f"File '{file_path}' does not exist.")

        # get logger
        logger = self.get_logger(logger)
        if isinstance(logger, ActionLogger):
            action_logger: ActionLogger = logger

        else:
            action_logger: ActionLogger = logger.action_logger
        action_log_download_url = f"{action_logger.admin.base_url}/action_log/download/{action_logger.action_log_key}"

        # get producer, exchange and download_queue
        producer: Producer = getattr(action_logger, "_producer")
        exchange: Exchange = getattr(action_logger, "_exchange")
        download_queue: Queue = getattr(action_logger, "_download_queue")

        # send download command
        producer.publish({"file_path": str(file_path)}, exchange=exchange, routing_key=download_queue.name)

        if icon is None:
            icon = "fa fa-download"
        if form is None:
            form = f"""<form>
                <a href="{action_log_download_url}" download>
                <p>{text}</p>
            </form>"""

        super().__init__(logger=logger,
                         text=text,
                         submit_btn_text=submit_btn_text,
                         form=form)

    def __call__(self, timeout: Optional[float] = None) -> None:
        super().__call__(timeout=timeout)


class YesNoCommand(FormCommand):
    def __init__(self,
                 logger: Union["ActionLogger", "ActionSubLogger", logging.Logger],
                 text: str, submit_btn_text: Optional[str] = None,
                 abort_btn_text: Optional[str] = None,
                 default_value: Optional[bool] = None,
                 form: Optional[str] = None):
        if form is None:
            form = f"""<form>
            <div class="mt-3">
                <p>{text}</p>
            </div>
            </form>"""
        if submit_btn_text is None:
            submit_btn_text = "Yes"
        if abort_btn_text is None:
            abort_btn_text = "No"
        super().__init__(logger=logger,
                         form=form,
                         submit_btn_text=submit_btn_text,
                         abort_btn_text=abort_btn_text,
                         default_values=default_value)

    def __call__(self, timeout: Optional[float] = None) -> bool:
        result = super().__call__(timeout=timeout)
        if type(result) is not bool:
            raise ValueError("Invalid response.")
        return result


class FinalizeCommand(_SubLoggerCommand):
    def __init__(self,
                 logger: Union["ActionSubLogger", logging.Logger],
                 success: bool,
                 on_success_msg: Optional[str] = None,
                 on_error_msg: Optional[str] = None,
                 on_error_msg_simple: Optional[str] = None,
                 end_steps: Optional[bool] = None):
        super().__init__(logger=logger,
                         allowed_logger_cls=[ActionSubLogger, logging.Logger],
                         command="finalize",
                         success=success,
                         on_success_msg=on_success_msg,
                         on_error_msg=on_error_msg,
                         on_error_msg_simple=on_error_msg_simple,
                         end_steps=end_steps)


class ExitCommand(_SubLoggerCommand):
    def __init__(self,
                 logger: Union["ActionSubLogger", logging.Logger]):
        super().__init__(logger=logger,
                         allowed_logger_cls=[ActionSubLogger, logging.Logger],
                         command="exit")


class ActionLoggerCommand(BaseModel):
    class Command(str, Enum):
        START = "start"
        INCREASE_STEPS = "increase_steps"
        STEP = "step"
        NEXT_STEP = "next_step"
        LOG = "log"
        FORM = "form"
        FINALIZE = "finalize"
        EXIT = "exit"

    sub_logger: str
    command: Command
    value: Union[str, int, dict[str, Any]]


class ActionLoggerResponse(BaseModel):
    class Command(str, Enum):
        FORM = "form"

    sub_logger: str
    command: Command
    value: dict[str, Any]


class WebsocketHandler(logging.Handler):
    def __init__(self, sub_logger: "ActionSubLogger"):
        """
        Create new websocket handler.

        :return: None
        """

        super().__init__()

        self.sub_logger: ActionSubLogger = sub_logger

    def emit(self, record: logging.LogRecord) -> None:
        """
        Emit log record.

        :param record: Log record
        :return: None
        """

        # get extra
        sub_logger_name = getattr(record, "sub_logger")
        command: dict[str, Any] = getattr(record, "command", None)

        command_dict = {"sub_logger": sub_logger_name}

        # check if record is command
        if command is not None:
            command_dict.update(command)
        else:
            msg = self.format(record)
            command_dict.update({"command": "log", "value": msg})

        # send command to ActionLogger
        self.sub_logger.action_logger.send_command(command_dict)


class ActionSubLogger(logging.Logger):
    def __init__(self,
                 action_logger: "ActionLogger",
                 name: str,
                 title: Optional[str] = None,
                 parent: Optional[logging.Logger] = None,
                 log_level: Optional[ActionLogAdminSettings.LogLevels] = None,
                 formatter: Optional[logging.Formatter] = None,
                 websocket_handler_cls: Optional[type[WebsocketHandler]] = None):
        """
        Create new action sub logger.

        :param action_logger: Action logger
        :param name: Name of sub logger. Only a-z, A-Z, 0-9, - and _ are allowed.
        :param title: Title of sub logger. Visible in frontend.
        :param websocket_handler_cls: Websocket handler class.
        """

        # validate name
        if not name:
            raise ValueError("Name must not be empty.")
        for char in name:
            if char not in string.ascii_letters + string.digits + "-" + "_":
                raise ValueError("Invalid character in name. Only a-z, A-Z, 0-9, - and _ are allowed.")

        super().__init__(name=action_logger.action_log_key + "." + name)

        self._action_logger = action_logger

        # set title
        if title is None:
            title = name
        self._title = title

        # set parent
        if parent is None:
            parent = self._action_logger.parent
        self.parent = parent

        # set log level
        if log_level is None:
            log_level = self._action_logger.log_level
        self.setLevel(log_level)

        # set formatter
        if formatter is None:
            formatter = self._action_logger.formatter
        self._formatter = formatter

        # initialize variables
        self._started: bool = False
        self._steps: Optional[int] = None
        self._step: int = 0
        self._error_occurred: bool = False
        self._finalize_msg: Optional[str] = None
        self._finalize_msg_simple: Optional[str] = None
        self._response_obj: Union[None, bool, ActionLoggerResponse] = None

        # check if logger already exists
        if self.is_logger_exist(name=self.name):
            raise ValueError("ActionSubLogger already exists.")

        # set websocket_handler_cls
        self.websocket_handler_cls: type[WebsocketHandler] = websocket_handler_cls or self._action_logger.websocket_handler_cls

    def __del__(self):
        if not self.exited:
            self.exit()

    @property
    def action_logger(self) -> "ActionLogger":
        """
        Get action logger.

        :return: Action logger.
        """

        return self._action_logger

    @classmethod
    def _get_logger(cls, name: str) -> Optional["ActionSubLogger"]:
        """
        Get logger by name.

        :param name: Name of logger.
        :return: Logger
        """

        # get all logger
        all_loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]

        # filter action logger
        for _logger in all_loggers:
            if name != _logger.name:
                continue
            if not isinstance(_logger, ActionSubLogger):
                continue
            return _logger
        return None

    @classmethod
    def is_logger_exist(cls, name: str) -> bool:
        """
        Check if logger exists by name.

        :param name: Name of logger.
        :return: True if exists, otherwise False.
        """

        return cls._get_logger(name=name) is not None

    def handle(self, record) -> None:
        record.sub_logger = self.sub_logger_name

        if hasattr(record, "handling_command"):
            command_name: str = getattr(record, "handling_command")["command"]
            values: dict[str, Any] = getattr(record, "handling_command")["values"]
            del record.handling_command

            command = {"command": command_name}
            if command_name == "start":
                if self.started:
                    raise ValueError("ActionSubLogger already started.")
                if not self.exited:
                    raise ValueError("ActionSubLogger not exited.")

                # add websocket handler
                self.addHandler(self.websocket_handler_cls(sub_logger=self))

                # set formatter
                for handler in self.handlers:
                    handler.setFormatter(self._formatter)

                # add logger to logger manager
                logging.root.manager.loggerDict[self.name] = self

                command["value"] = self.title
                record.command = command
                super().handle(record)
                self._started = True

                # set steps
                steps = values["steps"]
                if steps is not None:
                    self.steps = steps
            elif command_name == "step":
                if not self.started:
                    raise ValueError("ActionSubLogger not started.")
                if self.exited:
                    raise ValueError("ActionSubLogger already exited.")
                step = values["step"]
                steps = values["steps"]
                if steps is None:
                    return
                if steps < 0:
                    raise ValueError("Steps must be greater than 0.")
                if step >= steps:
                    step = steps
                calculated_progress = round(step / steps * 100)
                command["value"] = calculated_progress
                record.command = command
                super().handle(record)
                self._step = step
                self._steps = steps
            elif command_name == "next_step":
                if not self.started:
                    raise ValueError("ActionSubLogger not started.")
                if self.exited:
                    raise ValueError("ActionSubLogger already exited.")
                steps = values["steps"]
                self.step += steps
            elif command_name == "increase_steps":
                if not self.started:
                    raise ValueError("ActionSubLogger not started.")
                if self.exited:
                    raise ValueError("ActionSubLogger already exited.")
                steps = values["steps"]
                if steps < 0:
                    raise ValueError("Steps must be greater than 0.")
                if self.steps is None:
                    self.steps = steps
                else:
                    self.steps += steps
            elif command_name == "form":
                if not self.started:
                    raise ValueError("ActionSubLogger not started.")
                if self.exited:
                    raise ValueError("ActionSubLogger already exited.")
                submit_btn_text = values["submit_btn_text"]
                abort_btn_text = values["abort_btn_text"]
                form = values["form"]
                command["value"] = {"submit_btn_text": submit_btn_text, "abort_btn_text": abort_btn_text, "form": form}
                record.command = command
                super().handle(record)
            elif command_name == "finalize":
                if not self.started:
                    raise ValueError("ActionSubLogger not started.")
                if self.exited:
                    raise ValueError("ActionSubLogger already exited.")
                success = values["success"]
                end_steps = values["end_steps"]
                self._error_occurred = not success
                self._finalize_msg = values["on_success_msg"] if success else values["on_error_msg"]
                self._finalize_msg_simple = values["on_error_msg_simple"]
                contexts = ActionSubLoggerContext.get_from_stack(inspect.stack())
                current_context: Union[None, ActionSubLoggerContext, LoggingContext] = None
                if len(contexts) > 0:
                    current_context = contexts[-1]
                if not isinstance(current_context, ActionSubLoggerContext) and current_context is not None:
                    raise ValueError("Wrong context.")
                if end_steps is None:
                    if current_context is not None:
                        end_steps = current_context.end_steps

                if success:
                    if end_steps is None:
                        end_steps = True
                    if self._finalize_msg is None:
                        if current_context is not None:
                            self._finalize_msg = current_context.on_success_msg
                    if self._finalize_msg is None:
                        self._finalize_msg = "Success."
                    self.log(logging.INFO, self.finalize_msg)

                else:
                    if end_steps is None:
                        end_steps = False
                    if self._finalize_msg is None:
                        if current_context is not None:
                            self._finalize_msg = current_context.on_error_msg
                    if self._finalize_msg is None:
                        self._finalize_msg = "Something went wrong."
                    self.log(logging.ERROR, self.finalize_msg)

                if self.steps is not None and end_steps:
                    if self.step < self.steps:
                        self.step = self.steps

                command["value"] = success
                record.command = command
                super().handle(record)
                self.exit()
            elif command_name == "exit":
                if not self.started:
                    raise ValueError("ActionSubLogger not started.")
                if self.exited:
                    raise ValueError("ActionSubLogger already exited.")

                # remove handler
                for handler in self.handlers:
                    self.removeHandler(handler)

                # remove logger from logger manager
                logging.root.manager.loggerDict.pop(self.name, None)
            else:
                raise ValueError("Invalid command.")
        else:
            super().handle(record)

    @property
    def sub_logger_name(self) -> str:
        """
        Get sub logger name.

        :return: Sub logger name.
        """

        return self.name.replace(self._action_logger.action_log_key + ".", "")

    @property
    def title(self) -> str:
        """
        Get title of sub logger.

        :return: Title of sub logger.
        """

        return self._title

    def start(self, steps: Optional[int] = None) -> None:
        """
        Start sub logger.

        :param steps: Steps
        :return: None
        """

        StartCommand(logger=self, steps=steps)

    @property
    def started(self) -> bool:
        """
        Check if sub logger is started.

        :return: True if started, otherwise False.
        """

        return self._started

    @property
    def steps(self) -> int:
        """
        Get steps of sub logger.

        :return: Steps of sub logger.
        """

        return self._steps

    @steps.setter
    def steps(self, value: int) -> None:
        """
        Set steps of sub logger. Also send step command to websocket.

        :param value: Steps
        :return: None
        """

        StepCommand(logger=self, step=self.step, steps=value)

    @property
    def step(self) -> int:
        """
        Get step of sub logger.

        :return: Step of sub logger.
        """
        return self._step

    @step.setter
    def step(self, value: int) -> None:
        """
        Set step of sub logger. Also send step command to websocket.

        :param value: Step
        :return: None
        """

        StepCommand(logger=self, step=value, steps=self.steps)

    def next_step(self) -> None:
        """
        Increase step by 1.

        :return: None
        """

        NextStepCommand(logger=self)

    def form(self, form: str, submit_btn_text: Optional[str] = None, abort_btn_text: Optional[str] = None) -> FormCommand:
        """
        Send form to frontend.

        :param form: Form HTML.
        :param submit_btn_text: Text of submit button.
        :param abort_btn_text: Text of cancel button.
        :return: Form data.
        """

        return FormCommand(logger=self, form=form, submit_btn_text=submit_btn_text, abort_btn_text=abort_btn_text)

    def confirm(self, text: str, submit_btn_text: Optional[str] = None, form: Optional[str] = None) -> ConfirmCommand:
        """
        Send confirm form to frontend.

        :param text: Text of confirm form.
        :param submit_btn_text: Text of submit button.
        :param form: Form HTML.
        :return: Form data.
        """

        return ConfirmCommand(logger=self, text=text, submit_btn_text=submit_btn_text, form=form)

    def download(self, file_path: Union[str, Path], text: str, icon: Optional[str] = None, submit_btn_text: Optional[str] = None,
                 form: Optional[str] = None) -> DownloadCommand:
        """
        Send download form to frontend.

        :param file_path: File path
        :param text: Text of download form.
        :param icon: Icon of download form.
        :param submit_btn_text: Text of submit button.
        :param form: Form HTML.
        :return: Form data.
        """

        return DownloadCommand(logger=self, file_path=file_path, text=text, icon=icon, submit_btn_text=submit_btn_text, form=form)

    def yes_no(self, text: str, submit_btn_text: Optional[str] = None, abort_btn_text: Optional[str] = None, form: Optional[str] = None) -> YesNoCommand:
        """
        Send yes/no form to frontend.

        :param text: Text of yes/no form.
        :param submit_btn_text: Text of submit button.
        :param abort_btn_text: Text of cancel button.
        :param form: Form HTML.
        :return: Form data.
        """

        return YesNoCommand(logger=self, text=text, submit_btn_text=submit_btn_text, abort_btn_text=abort_btn_text, form=form)

    async def await_response(self, timeout: Optional[float] = None) -> Union[bool, ActionLoggerResponse]:
        """
        Fetch response from frontend.

        :param timeout: Timeout in seconds.
        :return: Form data.
        """

        return await self.action_logger._await_response(logger=self, timeout=timeout)

    async def form_data(self, timeout: Optional[float] = None) -> Union[bool, dict[str, Any]]:
        """
        Fetch form data from frontend.

        :param timeout: Timeout in seconds.
        :return: Form data.
        """

        return await self.action_logger._form_data(logger=self, timeout=timeout)

    @property
    def awaiting_response(self) -> bool:
        """
        Check if sub logger is awaiting response.

        :return: True if awaiting response, otherwise False.
        """

        return self.action_logger._awaiting_response(logger=self)

    def finalize(self,
                 success: bool = True,
                 on_success_msg: Optional[str] = None,
                 on_error_msg: Optional[str] = None,
                 on_error_msg_simple: Optional[str] = None,
                 end_steps: Optional[bool] = None) -> None:
        """
        Finalize sub logger. Also send finalize command to websocket.

        :param success: If True, frontend will show success message. If False, frontend will show error message.
        :param on_success_msg: Message if success.
        :param on_error_msg: Message if error.
        :param on_error_msg_simple: Simple message if error.
        :param end_steps: End steps on finalize.
        :return: None
        """

        FinalizeCommand(logger=self, success=success, on_success_msg=on_success_msg, on_error_msg=on_error_msg, on_error_msg_simple=on_error_msg_simple, end_steps=end_steps)

    def exit(self) -> None:
        """
        Exit sub logger. Also remove websocket from sub logger.

        :return: None
        """

        ExitCommand(logger=self)

    @property
    def exited(self) -> bool:
        """
        Check if sub logger is exited.

        :return: True if exited, otherwise False.
        """

        return not self.is_logger_exist(name=self.name)

    @property
    def finalize_msg(self) -> str:
        """
        Get finalize message.

        :return: Finalize message.
        """

        return self._finalize_msg

    @property
    def finalize_msg_simple(self) -> str:
        """
        Get finalize simple message.

        :return: Finalize simple message.
        """

        if self._finalize_msg_simple is None:
            return self._finalize_msg

        return self._finalize_msg_simple

    @property
    def error_occurred(self) -> bool:
        """
        Check if error occurred.

        :return: True if error occurred, otherwise False.
        """

        return self._error_occurred


class ActionSubLoggerContext(LoggingContext):
    def __init__(self,
                 action_logger: "ActionLogger",
                 name: str,
                 title: Optional[str] = None,
                 log_level: Optional[ActionLogAdminSettings.LogLevels] = None,
                 parent: Optional[logging.Logger] = None,
                 formatter: Optional[logging.Formatter] = None,
                 steps: Optional[int] = None,
                 on_success_msg: Optional[str] = None,
                 on_error_msg: Optional[str] = None,
                 end_steps: Optional[bool] = None,
                 show_errors: Optional[bool] = None,
                 halt_on_error: Optional[bool] = None,
                 use_context_logger_level: Optional[bool] = None,
                 use_context_logger_level_on_not_set: Optional[bool] = None,
                 ignore_loggers_equal: Optional[list[str]] = None,
                 ignore_loggers_like: Optional[list[str]] = None,
                 handle_origin_logger: Optional[bool] = None,
                 action_sub_logger_cls: Optional[type[ActionSubLogger]] = None,
                 websocket_handler_cls: Optional[type[WebsocketHandler]] = None):
        """
        Create new action sub logger context manager.

        :param action_logger: Action logger
        :param name: Name of sub logger. Only a-z, A-Z, 0-9, - and _ are allowed.
        :param title: Title of sub logger. Visible in frontend.
        :param log_level: Log level of sub logger. If None, ac
        :param parent: Parent logger. If None, action logger parent will be used.
        :param formatter: Formatter of sub logger. If None, action logger formatter will be used.
        :param steps: Steps of sub logger.
        :param on_success_msg: Message of finalize message if success.
        :param on_error_msg: Message of finalize message if error.
        :param end_steps: End steps on finalize.
        :param show_errors: Show errors in frontend. If None, action logger show_errors will be used.
        :param halt_on_error: Halt on error.
        :param use_context_logger_level: Use context logger level.
        :param use_context_logger_level_on_not_set: Use context logger level on not set.
        :param ignore_loggers_equal: Ignore loggers equal.
        :param ignore_loggers_like: Ignore loggers like.
        :param handle_origin_logger: Handle origin logger.
        :param action_sub_logger_cls: Action sub logger class.
        :param websocket_handler_cls: Websocket handler class.
        """

        # define action logger
        self._action_logger = action_logger

        # set steps
        self._steps = steps

        # set finalize messages
        self.on_success_msg = on_success_msg
        self.on_error_msg = on_error_msg

        # set end steps
        self.end_steps = end_steps

        # set show errors
        if show_errors is None:
            show_errors = action_logger.show_errors
        self.show_errors = show_errors

        # set halt on error
        if halt_on_error is None:
            halt_on_error = action_logger.halt_on_error
        self.halt_on_error = halt_on_error

        # create sub logger
        self.context_logger = self._action_logger.new_sub_logger(name=name,
                                                                 title=title,
                                                                 log_level=log_level,
                                                                 parent=parent,
                                                                 formatter=formatter,
                                                                 action_sub_logger_cls=action_sub_logger_cls,
                                                                 websocket_handler_cls=websocket_handler_cls)

        super().__init__(context_logger=self.context_logger,
                         use_context_logger_level=use_context_logger_level,
                         use_context_logger_level_on_not_set=use_context_logger_level_on_not_set,
                         ignore_loggers_equal=ignore_loggers_equal,
                         ignore_loggers_like=ignore_loggers_like,
                         handle_origin_logger=handle_origin_logger)

    def __enter__(self) -> "ActionSubLogger":
        super().__enter__()
        self.context_logger.start(steps=self._steps)

        return self.context_logger

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        if self.context_logger.exited:
            return False
        else:
            if exc_type is None:
                self.context_logger.finalize(success=True, on_success_msg=self.on_success_msg, on_error_msg=self.on_error_msg, end_steps=self.end_steps)
            else:
                on_error_msg_simple = None
                if exc_type is ActionFailed:
                    on_error_msg = exc_val.args[0]
                else:
                    on_error_msg = None
                    if self.show_errors:
                        # get exception string
                        on_error_msg = traceback.format_exc()
                        on_error_msg_simple = f"{exc_type.__name__}: {exc_val}"
                self.context_logger.finalize(success=False,
                                             on_success_msg=self.on_success_msg,
                                             on_error_msg=on_error_msg,
                                             on_error_msg_simple=on_error_msg_simple,
                                             end_steps=False)
            return exc_type is None or not self.halt_on_error


class ActionLogger:
    _action_loggers: list["ActionLogger"] = []

    def __init__(self,
                 request_or_websocket: Union[Request],
                 log_level: Optional[ActionLogAdminSettings.LogLevels] = None,
                 parent: Optional[logging.Logger] = None,
                 formatter: Union[None, str, logging.Formatter] = None,
                 wait_for_websocket: Optional[bool] = None,
                 wait_for_websocket_timeout: Optional[int] = None,
                 show_errors: Optional[bool] = None,
                 halt_on_error: Optional[bool] = None,
                 use_context_logger_level: Optional[bool] = None,
                 use_context_logger_level_on_not_set: Optional[bool] = None,
                 ignore_loggers_equal: Optional[list[str]] = None,
                 ignore_loggers_like: Optional[list[str]] = None,
                 handle_origin_logger: Optional[bool] = None,
                 action_sub_logger_cls: Optional[type[ActionSubLogger]] = None,
                 websocket_handler_cls: Optional[type[WebsocketHandler]] = None):
        """
        Create new action logger.

        :param request_or_websocket: request or websocket.
        :param log_level: Log level of action logger. If None, parent log level will be used. If parent is None, logging.INFO will be used.
        :param parent: Parent logger. If None, logger will be added to module logger.
        :param formatter: Formatter of action logger. If None, default formatter will be used.
        :param show_errors: Show errors in frontend.
        :param halt_on_error: Halt on error.
        :param wait_for_websocket: Wait for websocket to be connected. For this feature, await must be used.
        :param wait_for_websocket_timeout: Timeout in seconds. For this feature, await must be used.
        """

        # get action log key
        self.action_log_key = self.get_action_key(request_or_websocket=request_or_websocket)

        # get admin
        self.admin = request_or_websocket.app.state.admin

        # get settings
        self.settings: ActionLogAdminSettings = self.admin.settings

        # set log level
        if log_level is None:
            log_level = self.settings.action_log_level
        self.log_level: ActionLogAdminSettings.LogLevels = log_level

        # get parent logger
        if parent is None:
            parent = LOGGER
        self.parent: logging.Logger = parent

        # set formatter
        if formatter is None:
            formatter = self.settings.action_log_formatter
        self.formatter: logging.Formatter = logging.Formatter(formatter)

        # set wait for websocket
        if wait_for_websocket is None:
            wait_for_websocket = self.settings.action_log_wait_for_websocket
        self.wait_for_websocket: bool = wait_for_websocket

        # set wait for websocket timeout
        if wait_for_websocket_timeout is None:
            wait_for_websocket_timeout = self.settings.action_log_wait_for_websocket_timeout
        self.wait_for_websocket_timeout: int = wait_for_websocket_timeout

        # set show errors
        if show_errors is None:
            show_errors = self.settings.action_log_show_errors
        self.show_errors: bool = show_errors

        # set halt on error
        if halt_on_error is None:
            halt_on_error = self.settings.action_log_halt_on_error
        self.halt_on_error: bool = halt_on_error

        # set use context logger level
        if use_context_logger_level is None:
            use_context_logger_level = self.settings.action_log_use_context_logger_level
        self.use_context_logger_level: bool = use_context_logger_level

        # set use context logger level on not set
        if use_context_logger_level_on_not_set is None:
            use_context_logger_level_on_not_set = self.settings.action_log_use_context_logger_level_on_not_set
        self.use_context_logger_level_on_not_set: bool = use_context_logger_level_on_not_set

        # set ignore loggers equal
        if ignore_loggers_equal is None:
            ignore_loggers_equal = self.settings.action_log_ignore_loggers_equal
        self.ignore_loggers_equal: list[str] = ignore_loggers_equal

        # set ignore loggers like
        if ignore_loggers_like is None:
            ignore_loggers_like = self.settings.action_log_ignore_loggers_like
        self.ignore_loggers_like: list[str] = ignore_loggers_like
        if "pymongo" not in self.ignore_loggers_like:
            self.ignore_loggers_like.append("pymongo")  # force ignore loggers like 'pymongo'

        # set handle origin logger
        if handle_origin_logger is None:
            handle_origin_logger = self.settings.action_log_handle_origin_logger
        self.handle_origin_logger: bool = handle_origin_logger

        # set action sub logger class
        self.action_sub_logger_cls = action_sub_logger_cls or ActionSubLogger

        # set websocket handler class
        self.websocket_handler_cls = websocket_handler_cls or WebsocketHandler

        self._sub_logger: list[ActionSubLogger] = []
        self._response_obj: Union[None, bool, ActionLoggerResponse] = None

        # add action logger to action loggers
        self._action_loggers.append(self)

        # get kombu connection
        self._kombu_connection = self.get_kombu_connection(request_or_websocket=request_or_websocket)

        # create exchange and queues from websocket request
        self._exchange, self._start_queue, self._log_queue, self._response_queue, self._download_queue, self._exit_queue = ActionLogger.get_action_log_queues(
            request_or_websocket=request_or_websocket)

        # create producer
        self._producer = self._kombu_connection.Producer(serializer='json')

        # create exit thread
        self._exit_thread_obj = Thread(target=self._exit_thread)

    def __await__(self):
        async def _await() -> "ActionLogger":
            if not self.wait_for_websocket:
                return self

            current_try = 0
            connected = False

            def start_event(body: dict[str, Any], message: Message):
                nonlocal connected
                connected = True

                message.ack()

            with self._kombu_connection.Consumer([self._start_queue], callbacks=[start_event]):
                while len(self._sub_logger) == 0:
                    if connected:
                        break
                    if current_try >= self.wait_for_websocket_timeout:
                        self.exit()
                        raise ValueError("No websocket connected.")

                    current_try += 1
                    LOGGER.debug(f"[{current_try}/{self.wait_for_websocket_timeout}] Waiting for websocket...")
                    await asyncio.sleep(1)

            return self

        return _await().__await__()

    def __enter__(self) -> "ActionLogger":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.exited:
            self.exit()

        # get exception string
        if exc_type is not None and self.show_errors:
            te = traceback.TracebackException(type(exc_type), exc_val, exc_tb)
            efs = te.stack[-1]
            exception_str = f"{exc_type.__name__}: {exc_val}"
            # add line number
            if exc_tb is not None:
                exception_str += f" at line {efs.lineno} in {efs.filename}"

            # raise ActionFailed
            raise ActionFailed(exception_str)

        # check if error occurred in sub logger
        finalize_msg = ""
        for sub_logger in self._sub_logger:
            if sub_logger.error_occurred:
                finalize_msg += f"{sub_logger.title}: {sub_logger.finalize_msg_simple}\n"
        if finalize_msg:
            raise ActionFailed(finalize_msg)

    def __del__(self):
        if not self.exited:
            self.exit()

    @classmethod
    def get_action_key(cls, request_or_websocket: Union[Request, WebSocket]) -> str:
        """
        Get action log key from request or websocket.

        :param request_or_websocket: Action log key, request or websocket.
        :return: Action log key.
        """

        if isinstance(request_or_websocket, Request):
            action_log_key = request_or_websocket.query_params.get("action_log_key", None)
            if action_log_key is None:
                action_log_key = request_or_websocket.path_params.get("action_log_key", None)
            if action_log_key is None:
                raise ValueError("No action log key provided.")
        elif isinstance(request_or_websocket, WebSocket):
            action_log_key = request_or_websocket.path_params.get("action_log_key", None)
            if action_log_key is None:
                raise ValueError("No action log key provided.")
        else:
            raise ValueError("Invalid action log key or request.")
        return action_log_key

    @classmethod
    def get_kombu_connection(cls, request_or_websocket: Union[Request, WebSocket]) -> Connection:
        return request_or_websocket.app.state.admin.kombu_connection

    @classmethod
    def get_action_log_exchange(cls, request_or_websocket: Union[Request, WebSocket]) -> Exchange:
        exchange = Exchange(KOMBU_EXCHANGE_NAME, "direct", durable=True)
        exchange = exchange.bind(cls.get_kombu_connection(request_or_websocket))
        exchange.declare()
        return exchange

    @classmethod
    def get_action_log_queues(cls, request_or_websocket: Union[Request, WebSocket]) -> tuple[Exchange, Queue, Queue, Queue, Queue, Queue]:
        # get kombu connection
        connection = cls.get_kombu_connection(request_or_websocket=request_or_websocket)

        # get action log exchange
        exchange = cls.get_action_log_exchange(request_or_websocket=request_or_websocket)

        # get action log key
        action_log_key = cls.get_action_key(request_or_websocket)

        # create exit queue
        start_queue_name = f"{action_log_key}_start"
        start_queue = Queue(start_queue_name, exchange=exchange, routing_key=start_queue_name)
        start_queue = start_queue.bind(connection)
        start_queue.declare()

        # create log queue
        log_queue_name = f"{action_log_key}_log"
        log_queue = Queue(log_queue_name, exchange=exchange, routing_key=log_queue_name)
        log_queue = log_queue.bind(connection)
        log_queue.declare()

        # create response queue
        response_queue_name = f"{action_log_key}_response"
        response_queue = Queue(response_queue_name, exchange=exchange, routing_key=response_queue_name)
        response_queue = response_queue.bind(connection)
        response_queue.declare()

        # create download queue
        download_queue_name = f"{action_log_key}_download"
        download_queue = Queue(download_queue_name, exchange=exchange, routing_key=download_queue_name)
        download_queue = download_queue.bind(connection)
        download_queue.declare()

        # create exit queue
        exit_queue_name = f"{action_log_key}_exit"
        exit_queue = Queue(exit_queue_name, exchange=exchange, routing_key=exit_queue_name)
        exit_queue = exit_queue.bind(connection)
        exit_queue.declare()

        return exchange, start_queue, log_queue, response_queue, download_queue, exit_queue

    @classmethod
    def parse_response_obj(cls, data: Union[str, dict[str, Any]]) -> Union[None, ActionLoggerResponse]:
        """
        Parse response object.

        :param data: Data
        :return: Response object
        """

        # parse to dict
        if type(data) is str:
            try:
                data = json.loads(data)
            except json.JSONDecodeError as e:
                LOGGER.error(f"JSONDecodeError while parsing response object: {e}")
                return None

        # create response object
        try:
            response_obj = ActionLoggerResponse(**data)
        except ValidationError as e:
            LOGGER.error(f"ValidationError while parsing response object: {e}")
            return None

        return response_obj

    def new_sub_logger(self,
                       name: str,
                       title: Optional[str] = None,
                       log_level: Optional[ActionLogAdminSettings.LogLevels] = None,
                       parent: Optional[logging.Logger] = None,
                       formatter: Optional[logging.Formatter] = None,
                       action_sub_logger_cls: Optional[type[ActionSubLogger]] = None,
                       websocket_handler_cls: Optional[type[WebsocketHandler]] = None) -> ActionSubLogger:
        """
        Create new sub logger.

        :param name: Name of sub logger. Only a-z, A-Z, 0-9, - and _ are allowed.
        :param title: Title of sub logger. Visible in frontend.
        :param log_level: Log level of sub logger. If None, action logger log level will be used.
        :param parent: Parent logger. If None, action logger parent will be used.
        :param formatter: Formatter of sub logger. If None, action logger formatter will be used.
        :param action_sub_logger_cls: Action sub logger class.
        :param websocket_handler_cls: Websocket handler class.
        :return: Sub logger.
        """

        try:
            self.get_sub_logger(sub_logger_name=name)
        except ValueError:
            pass

        # create sub logger
        action_sub_logger_cls = action_sub_logger_cls or self.action_sub_logger_cls
        sub_logger = action_sub_logger_cls(action_logger=self,
                                           name=name,
                                           title=title,
                                           parent=parent,
                                           log_level=log_level,
                                           formatter=formatter,
                                           websocket_handler_cls=websocket_handler_cls)

        self._sub_logger.append(sub_logger)
        return sub_logger

    def get_sub_logger(self, sub_logger_name: str) -> ActionSubLogger:
        """
        Get sub logger by name.

        :param sub_logger_name: Name of sub logger.
        :return:
        """

        if self.exited:
            raise ValueError("ActionLogger already exited.")

        # check if sub logger already exists
        for sub_logger in self._sub_logger:
            if sub_logger.sub_logger_name == sub_logger_name:
                return sub_logger
        raise ValueError("Sub logger not found.")

    def sub_logger(self,
                   name: str,
                   title: Optional[str] = None,
                   log_level: Optional[ActionLogAdminSettings.LogLevels] = None,
                   parent: Optional[logging.Logger] = None,
                   formatter: Optional[logging.Formatter] = None,
                   steps: Optional[int] = None,
                   on_success_msg: Optional[str] = None,
                   on_error_msg: Optional[str] = None,
                   end_steps: Optional[bool] = None,
                   show_errors: Optional[bool] = None,
                   halt_on_error: Optional[bool] = None,
                   use_context_logger_level: bool = True,
                   use_context_logger_level_on_not_set: Optional[bool] = None,
                   ignore_loggers_equal: Optional[list[str]] = None,
                   ignore_loggers_like: Optional[list[str]] = None,
                   handle_origin_logger: bool = True,
                   action_sub_logger_cls: Optional[type[ActionSubLogger]] = None,
                   websocket_handler_cls: Optional[type[WebsocketHandler]] = None) -> ActionSubLoggerContext:

        """
        Sub logger context manager.

        :param name: Name of sub logger. Only a-z, A-Z, 0-9, - and _ are allowed.
        :param title: Title of sub logger. Visible in frontend.
        :param log_level: Log level of sub logger. If None, parent log level will be used. If parent is None, action logger log level will be used.
        :param parent: Parent logger. If None, action logger parent will be used.
        :param formatter: Formatter of sub logger. If None, action logger formatter will be used.
        :param steps: Steps of sub logger.
        :param on_success_msg: Message of finalize message if success.
        :param on_error_msg: Message of finalize message if error.
        :param end_steps: End steps on finalize.
        :param show_errors: Show errors in frontend. If None, action logger show_errors will be used.
        :param halt_on_error: Halt on error.
        :param use_context_logger_level: Use context logger level.
        :param use_context_logger_level_on_not_set: Use context logger level on not set.
        :param ignore_loggers_equal: Ignore loggers equal to this list.
        :param ignore_loggers_like: Ignore loggers like this list.
        :param handle_origin_logger: Handle origin logger.
        :param action_sub_logger_cls: Action sub logger class.
        :param websocket_handler_cls: Websocket handler class.
        :return:
        """

        return ActionSubLoggerContext(action_logger=self,
                                      name=name,
                                      title=title,
                                      log_level=log_level,
                                      parent=parent,
                                      formatter=formatter,
                                      steps=steps,
                                      on_success_msg=on_success_msg,
                                      on_error_msg=on_error_msg,
                                      end_steps=end_steps,
                                      show_errors=show_errors,
                                      halt_on_error=halt_on_error,
                                      use_context_logger_level=use_context_logger_level,
                                      use_context_logger_level_on_not_set=use_context_logger_level_on_not_set,
                                      ignore_loggers_equal=ignore_loggers_equal,
                                      ignore_loggers_like=ignore_loggers_like,
                                      handle_origin_logger=handle_origin_logger,
                                      action_sub_logger_cls=action_sub_logger_cls,
                                      websocket_handler_cls=websocket_handler_cls)

    def form(self, form: str, submit_btn_text: Optional[str] = None, abort_btn_text: Optional[str] = None) -> FormCommand:
        """
        Send form to frontend.

        :param form: Form HTML.
        :param submit_btn_text: Text of submit button.
        :param abort_btn_text: Text of cancel button.
        :return: Form data.
        """

        return FormCommand(logger=self, form=form, submit_btn_text=submit_btn_text, abort_btn_text=abort_btn_text)

    def confirm(self, text: str, submit_btn_text: Optional[str] = None, form: Optional[str] = None) -> ConfirmCommand:
        """
        Send confirm form to frontend.

        :param text: Text of confirm form.
        :param submit_btn_text: Text of submit button.
        :param form: Form HTML.
        :return: Form data.
        """

        return ConfirmCommand(logger=self, text=text, submit_btn_text=submit_btn_text, form=form)

    def download(self, file_path: Union[str, Path], text: str, icon: Optional[str] = None, submit_btn_text: Optional[str] = None,
                 form: Optional[str] = None) -> DownloadCommand:
        """
        Send download form to frontend.

        :param file_path: File path
        :param text: Text of download form.
        :param icon: Icon of download form.
        :param submit_btn_text: Text of submit button.
        :param form: Form HTML.
        :return: Form data.
        """

        return DownloadCommand(logger=self, file_path=file_path, text=text, icon=icon, submit_btn_text=submit_btn_text, form=form)

    def yes_no(self, text: str, submit_btn_text: Optional[str] = None, abort_btn_text: Optional[str] = None, form: Optional[str] = None) -> YesNoCommand:
        """
        Send yes/no form to frontend.

        :param text: Text of yes/no form.
        :param submit_btn_text: Text of submit button.
        :param abort_btn_text: Text of cancel button.
        :param form: Form HTML.
        :return: Form data.
        """

        return YesNoCommand(logger=self, text=text, submit_btn_text=submit_btn_text, abort_btn_text=abort_btn_text, form=form)

    @classmethod
    async def _form_data(cls, logger: Union["ActionLogger", ActionSubLogger], timeout: Optional[float] = None) -> Union[bool, dict[str, Any]]:
        response_obj = await logger.await_response(timeout=timeout)
        if response_obj is False:
            return False

        # check if response object is form
        if response_obj.command != ActionLoggerResponse.Command.FORM:
            raise ValueError("Response object is not form.")

        # check "result" key is bool
        if not isinstance(response_obj.value["result"], bool):
            raise ValueError("Invalid result.")

        if len(response_obj.value["form_data"]) == 0 or not response_obj.value["result"]:
            return response_obj.value["result"]

        return response_obj.value["form_data"]

    async def form_data(self, timeout: Optional[float] = None) -> Union[bool, dict[str, Any]]:
        """
        Fetch form data from frontend.

        :param timeout: Timeout in seconds.
        :return: Form data.
        """

        return await self._form_data(logger=self, timeout=timeout)

    @classmethod
    def _awaiting_response(cls, logger: Union["ActionLogger", ActionSubLogger]) -> bool:
        if not isinstance(logger, ActionLogger) and not isinstance(logger, ActionSubLogger):
            raise ValueError("Invalid logger.")

        # get response object
        response_obj = getattr(logger, "_response_obj")

        if response_obj is None:
            return False
        elif type(response_obj) == bool:
            return True
        elif isinstance(response_obj, ActionLoggerResponse):
            return False
        else:
            raise ValueError("Invalid response object.")

    @property
    def awaiting_response(self) -> bool:
        """
        Check if logger is awaiting response.

        :return: True if awaiting response, otherwise False.
        """

        return self._awaiting_response(logger=self)

    def exit(self):
        """
        Exit action logger. Also remove all websockets and sub loggers.

        :return: None
        """

        if self.exited:
            raise ValueError("ActionLogger already exited.")

        # exit sub loggers
        for sub_logger in self._sub_logger:
            if not sub_logger.exited:
                sub_logger.exit()

        # remove action logger from action loggers
        self._action_loggers.remove(self)

        # exit thread
        self._exit_thread_obj.start()

    def _exit_thread(self):
        """
        Exit thread.

        :return: None
        """

        exited = False

        def exit_event(body: dict[str, Any], message: Message):
            nonlocal exited

            exited = True
            message.ack()

        with self._kombu_connection.Consumer([self._exit_queue], callbacks=[exit_event]):
            while not exited:
                try:
                    self._kombu_connection.drain_events(timeout=5)
                except socket_timeout:
                    if not exited:
                        warn(f"Exiting action logger {self.action_log_key} timed out.", UserWarning)
                        exited = True
                if exited:
                    break

        # delete queues
        self._start_queue.delete()
        self._log_queue.delete()
        self._response_queue.delete()
        self._download_queue.delete()
        self._exit_queue.delete()

    @property
    def exited(self) -> bool:
        """
        Check if action logger is exited.

        :return: True if exited, otherwise False.
        """

        return self not in self._action_loggers

    def send_command(self, command: Union[dict[str, Any], ActionLoggerCommand]) -> None:
        """
        Send command to all websockets.

        :param command: Command. If dict, it will be converted to ActionLoggerCommand.
        :return: None
        """

        # validate command
        if type(command) is dict:
            command = ActionLoggerCommand(**command)
        if not isinstance(command, ActionLoggerCommand):
            raise ValueError("Invalid command.")

        # convert command to dict
        command_dict = command.model_dump()

        # get action logger or sub logger
        if command.sub_logger == "":
            logger = self
        else:
            logger = self.get_sub_logger(sub_logger_name=command.sub_logger)

        # set _response_obj to True if command is available in ActionLoggerResponse.Command
        try:
            ActionLoggerResponse.Command(command.command.value)
            logger._response_obj = True
        except ValueError:
            ...

        self._producer.publish(command_dict, exchange=self._exchange, routing_key=self._log_queue.name)

    @classmethod
    async def _await_response(cls, logger: Union["ActionLogger", ActionSubLogger], timeout: Optional[float] = None) -> Union[bool, ActionLoggerResponse]:
        if not isinstance(logger, ActionLogger) and not isinstance(logger, ActionSubLogger):
            raise ValueError("Invalid logger.")

        if not logger.awaiting_response:
            raise ValueError("Logger is not awaiting form data.")

        action_logger = logger if isinstance(logger, ActionLogger) else logger.action_logger

        # wait for response
        start_wait = time.perf_counter()
        with action_logger._kombu_connection.Consumer([action_logger._response_queue], callbacks=[action_logger.send_response_to_logger]):
            while logger.awaiting_response:
                try:
                    action_logger._kombu_connection.drain_events(timeout=0.001)
                except socket_timeout:
                    ...
                await asyncio.sleep(0.001)
                if timeout is None:
                    continue
                if time.perf_counter() - start_wait >= timeout:
                    return False

        # get response object
        response_obj = getattr(logger, "_response_obj")
        setattr(logger, "_response_obj", None)
        logger_name = getattr(logger, "sub_logger_name", "")

        # check is response object is for this logger
        if not response_obj.sub_logger == logger_name:
            raise ValueError("The response object is not for this logger.")

        return response_obj

    async def await_response(self, timeout: Optional[float] = None) -> Union[bool, ActionLoggerResponse]:
        """
        Fetch response from frontend.

        :param timeout: Timeout in seconds.
        :return: Form data.
        """

        return await self._await_response(logger=self, timeout=timeout)

    def send_response_to_logger(self, body: dict[str, Any], message: Message) -> None:
        """
        Send response object to logger or sub logger.

        :param body: Body
        :param message: Message
        :return: None
        """

        message.ack()

        response_obj = self.parse_response_obj(body)

        if response_obj.sub_logger == "":
            logger = self
        else:
            # get sub logger
            logger = self.get_sub_logger(sub_logger_name=response_obj.sub_logger)

        # check if sub logger is awaiting response
        if not logger.awaiting_response:
            raise ValueError("Logger is not awaiting response.")

        # set response object
        setattr(logger, "_response_obj", response_obj)



from typing import Any, Dict, Type, Optional
import mongoengine as me

from starlette.requests import Request
from starlette_admin import action
from starlette_admin import RequestAction
from starlette_admin.contrib.mongoengine import ModelView as BaseModelView
from starlette_admin.contrib.mongoengine.converters import BaseMongoEngineModelConverter

from wiederverwendbar.starlette_admin.mongoengine.generic_embedded_document_field.view import GenericEmbeddedDocumentView
from wiederverwendbar.starlette_admin.mongoengine.converter import MongoengineConverter


class FixedModelView(BaseModelView):
    async def serialize(
            self,
            obj: Any,
            request: Request,
            action: RequestAction,
            include_relationships: bool = True,
            include_select2: bool = False,
    ) -> Dict[str, Any]:
        result = await super().serialize(obj, request, action, include_relationships, include_select2)
        result["id"] = str(result["id"])
        return result


class MongoengineModelView(FixedModelView, GenericEmbeddedDocumentView):
    def __init__(
            self,
            document: Type[me.Document],
            icon: Optional[str] = None,
            name: Optional[str] = None,
            label: Optional[str] = None,
            identity: Optional[str] = None,
            converter: Optional[BaseMongoEngineModelConverter] = None,
    ):
        super().__init__(document=document, icon=icon, name=name, label=label, identity=identity, converter=converter or MongoengineConverter())



from typing import Type, Optional, Callable, Union, Sequence, Tuple, Any

from starlette.requests import Request
from starlette_admin.contrib.mongoengine.converters import BaseMongoEngineModelConverter
from starlette_admin.fields import PasswordField, EnumField

from wiederverwendbar.starlette_admin.mongoengine.auth.documents.user import User
from wiederverwendbar.starlette_admin.mongoengine.view import MongoengineModelView


class UserView(MongoengineModelView):
    exclude_fields_from_list = [User.id,
                                "password",
                                User.sessions,
                                User.company_logo]
    exclude_fields_from_detail = [User.id,
                                  "password"]
    exclude_fields_from_create = [User.id,
                                  User.password_change_time,
                                  User.sessions]
    exclude_fields_from_edit = [User.id,
                                User.password_change_time,
                                User.sessions]

    def __init__(
            self,
            document: Type[User],
            icon: Optional[str] = None,
            name: Optional[str] = None,
            label: Optional[str] = None,
            identity: Optional[str] = None,
            converter: Optional[BaseMongoEngineModelConverter] = None,
            company_logo_choices_loader: Optional[Callable[[Request], Union[Sequence[str], Sequence[Tuple[Any, str]]]]] = lambda _: [],
    ):
        # set default values
        document = document or User
        icon = icon or "fa fa-user"
        name = name or "User"
        label = label or "Benutzer"

        fields = []
        for field_name in list(getattr(document, "_fields_ordered", [])):
            if field_name == "password_doc":
                fields.append(PasswordField(name="password"))
            elif field_name == "company_logo":
                fields.append(EnumField(name=field_name, choices_loader=company_logo_choices_loader))
            else:
                fields.append(field_name)
        self.fields = fields

        super().__init__(document=document,
                         icon=icon,
                         name=name,
                         label=label,
                         identity=identity,
                         converter=converter)
        for field in self.fields:
            if field.name == "username":
                field.label = "Benutzername"
            elif field.name == "password":
                field.label = "Passwort"
            elif field.name == "password_change_time":
                field.label = "Passwortänderungszeit"
            elif field.name == "password_expiration_time":
                field.label = "Passwortablaufzeit"
            elif field.name == "avatar":
                field.label = "Profilbild"
            elif field.name == "company_logo":
                field.label = "Firmenlogo"



import os
import hashlib
from typing import Any, Optional

ENCODING = "utf-8"
HASH_FUNCTION = "sha256"
INTERACTIONS = 100000
KEY_LENGTH = 128
SALT_LENGTH = 32


def hash_password(password: str,
                  encoding: Optional[str] = None,
                  hash_function: Optional[str] = None,
                  interactions: Optional[int] = None,
                  key_length: Optional[int] = None,
                  salt: Optional[str] = None) -> dict[str, Any]:
    encoding = encoding or ENCODING
    hash_function = hash_function or HASH_FUNCTION
    interactions = interactions or INTERACTIONS
    key_length = key_length or KEY_LENGTH
    salt = salt or os.urandom(SALT_LENGTH).hex()

    salt_encoded = salt.encode(encoding)
    password_encoded = password.encode(encoding)
    hashed_password_encoded = hashlib.pbkdf2_hmac(hash_function, password_encoded, salt_encoded, interactions, key_length)  # generate hash
    hashed_password = hashed_password_encoded.hex()
    out = {
        "encoding": encoding,
        "hash_function": hash_function,
        "interactions": interactions,
        "key_length": key_length,
        "salt": salt,
        "hashed_password": hashed_password
    }
    return out


def verify_password(hashed_password: dict[str, Any], verifying_password: str) -> bool:
    hashed_verify_password = hash_password(
        password=verifying_password,
        encoding=hashed_password["encoding"],
        hash_function=hashed_password["hash_function"],
        interactions=hashed_password["interactions"],
        key_length=hashed_password["key_length"],
        salt=hashed_password["salt"]
    )
    result = hashed_password["hashed_password"] == hashed_verify_password["hashed_password"]
    return result


class HashedPassword:
    def __init__(self,
                 encoding: str,
                 hash_function: str,
                 interactions: int,
                 key_length: int,
                 salt: str,
                 hashed_password: str):
        self.encoding = encoding
        self.hash_function = hash_function
        self.interactions = interactions
        self.key_length = key_length
        self.salt = salt
        self.hashed_password = hashed_password

    @classmethod
    def hash_password(cls,
                      password: str,
                      encoding: Optional[str] = None,
                      hash_function: Optional[str] = None,
                      interactions: Optional[int] = None,
                      key_length: Optional[int] = None,
                      salt: Optional[str] = None) -> Any:
        # create hashed password dict
        hashed_password = hash_password(
            password=password,
            encoding=encoding,
            hash_function=hash_function,
            interactions=interactions,
            key_length=key_length,
            salt=salt
        )

        # create hashed password object
        hashed_password_obj = cls(**hashed_password)

        return hashed_password_obj

    def verify_password(self, verifying_password: str) -> bool:
        hashed_password = {
            "encoding": self.encoding,
            "hash_function": self.hash_function,
            "interactions": self.interactions,
            "key_length": self.key_length,
            "salt": self.salt,
            "hashed_password": self.hashed_password
        }
        result = verify_password(hashed_password, verifying_password)
        return result



import typing
import warnings
import logging
from pathlib import Path
from typing import Optional, Union, Literal

logger = logging.getLogger(__name__)


# define warning ...
class HashTypeIsNoneWarning(Warning):
    pass


SUPPORTED_HASH_TYPES = Literal["md5", "sha1", "sha256", "sha512"]
SUPPORTED_HASH_TYPES_TUPLE: tuple[SUPPORTED_HASH_TYPES, ...] = typing.get_args(SUPPORTED_HASH_TYPES)


def check_hash_file(check_file: Union[str, Path],
                    hash_value: str,
                    hash_type: Optional[SUPPORTED_HASH_TYPES] = None) -> bool:
    """
    Check single file with hash value.

    :param check_file: file on disk
    :param hash_value: hash value for file
    :param hash_type: hash type for file, if None all hash types are checked(slows down the check)
    :return: True on success, False on fail
    """

    if hash_type is None:
        warnings.warn("No hash type specified, all hash types are checked. Define hash_type for better performance or suppress this warning.", HashTypeIsNoneWarning)

    if type(check_file) == str:
        check_file = Path(check_file)

    logger.debug(f"Check file: {check_file}")

    # check if file exists
    if not check_file.is_file():
        raise FileNotFoundError("File not found")

    def check_hash(ht: SUPPORTED_HASH_TYPES) -> bool:
        if ht == "md5":
            import hashlib
            h = hashlib.md5()
        elif ht == "sha1":
            import hashlib
            h = hashlib.sha1()
        elif ht == "sha256":
            import hashlib
            h = hashlib.sha256()
        elif ht == "sha512":
            import hashlib
            h = hashlib.sha512()
        else:
            raise ValueError("Unknown hash type")
        with open(check_file, "rb") as f:
            while True:
                data = f.read(1024)
                if not data:
                    break
                h.update(data)
        result = h.hexdigest() == hash_value
        logger.debug(f"Check hash: {ht} -> {result}")
        return result

    # check if hash_type is None
    if hash_type is None:
        for ht in SUPPORTED_HASH_TYPES_TUPLE:
            if check_hash(ht):
                return True
    else:
        if check_hash(hash_type):
            return True
    return False



from wiederverwendbar.mongoengine.logger.documets import (MongoengineLogDocument)
from wiederverwendbar.mongoengine.logger.formatters import (MongoengineLogFormatter)
from wiederverwendbar.mongoengine.logger.handlers import (MongoengineLogHandler)
from wiederverwendbar.mongoengine.logger.streamer import (MongoengineLogStreamer,
                                                          mongoengine_log_stream_print)



from typing import Union, Any

from mongoengine import BooleanField

from wiederverwendbar.mongoengine.fields.with_instance_field import WithInstanceField


class BooleanAlsoField(WithInstanceField):
    def __init__(self, also: Union[str, Any] = None, **kwargs):
        super().__init__(**kwargs)
        self.also: BooleanField = also

    def _set_owner_document(self, owner_document):
        super()._set_owner_document(owner_document)
        if type(self.also) is str:
            self.also = getattr(self.owner_document, self.also)
        if self.also is not None and not (isinstance(self.also, BooleanField) or isinstance(self.also, BooleanAlsoField)):
            raise ValueError("The field 'also' must be of type 'BooleanField'.")

    def validate(self, value, clean=True):
        # validate value
        if not isinstance(value, bool):
            self.error("BooleanField only accepts boolean values")

        # skip false values
        if not value:
            return

        # check if also field is set
        if self.also is None:
            return

        # get instance
        instance = self.get_instance()

        # set value of also field
        setattr(instance, self.also.name, value)

        # trigger validation of also field
        self.also.validate(value)

    def to_python(self, value):
        try:
            value = bool(value)
        except (ValueError, TypeError):
            pass
        return value



import time

from tests.shared import manager

if __name__ == '__main__':
    worker = manager.create_worker()

    while True:
        try:
            time.sleep(1)
        except KeyboardInterrupt:
            break
    # raise_exception = Exception("hallo welt")
    # worker.stop()
    # thread.raise_exception(Exception("hallo welt"))



from pathlib import Path

from sqlalchemy import Column, Text, Integer

from wiederverwendbar.logger import LoggerSingleton, LoggerSettings
from wiederverwendbar.sqlalchemy import Base, SqlalchemySettings, SqlalchemyDbSingleton

LoggerSingleton(name="test", settings=LoggerSettings(log_level=LoggerSettings.LogLevels.DEBUG), init=True)


class MyBase(Base, SqlalchemyDbSingleton(settings=SqlalchemySettings(db_file=Path("test.db")), init=True).Base):
    __abstract__ = True


class MyTable(MyBase):
    __tablename__ = "my_table"
    __str_columns__: list[str] = ["name"]

    name = Column(Text(50), primary_key=True)
    value = Column(Integer)


if __name__ == '__main__':
    SqlalchemyDbSingleton().create_all()

    my_table = MyTable(name="test", value=1)

    d = my_table.as_dict()
    my_table.save()

    print()



import atexit
import threading
import time
from datetime import datetime
from typing import Optional

from wiederverwendbar.functions.datetime import local_now
from wiederverwendbar.mongoengine.logger.documets import MongoengineLogDocument


class MongoengineLogStreamer(threading.Thread):
    def __init__(self,
                 log_document: type[MongoengineLogDocument],
                 search: Optional[dict] = None,
                 to: Optional[callable] = None,
                 name: Optional[str] = None,
                 begin: Optional[datetime] = None,
                 stream_rate: Optional[float] = None):
        if name is None:
            name = self.__class__.__name__
        super().__init__(name=name, daemon=True)
        self._lock = threading.Lock()

        if not issubclass(log_document, MongoengineLogDocument):
            raise TypeError(f"Expected '{MongoengineLogDocument.__name__}', got '{log_document}'.")
        self._log_document = log_document
        if search is None:
            search = {}
        if not isinstance(search, dict):
            raise TypeError(f"Expected 'dict', got '{type(search)}'.")
        self._search = search
        if to is None:
            to = mongoengine_log_stream_print
        if not callable(to):
            raise TypeError(f"Expected 'callable', got '{type(to)}'.")
        self._to = to
        if begin is None:
            begin = local_now()
        self._timestamp = begin
        if stream_rate is None:
            stream_rate = 0.001
        self._stream_rate = stream_rate
        self._buffer: list[dict] = []
        self._stopper = threading.Event()

        atexit.register(self.close)

    def __del__(self):
        self.close()

    def __enter__(self) -> 'MongoengineLogStreamer':
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def timestamp(self) -> datetime:
        with self._lock:
            return self._timestamp

    @timestamp.setter
    def timestamp(self, value: datetime):
        with self._lock:
            self._timestamp = value

    @property
    def stream_rate(self) -> float:
        with self._lock:
            return self._stream_rate

    @stream_rate.setter
    def stream_rate(self, value: float):
        with self._lock:
            self._stream_rate = value

    def _fetch(self) -> bool:
        entries = []
        for log_document in self._log_document.objects(timestamp__gt=self.timestamp, **self._search):
            log_document: MongoengineLogDocument
            for entry in log_document.entries:
                entries.append(entry)
            self.timestamp = log_document.timestamp
        with self._lock:
            self._buffer.extend(entries)
            return bool(self._buffer)

    def _stream(self):
        with self._lock:
            while self._buffer:
                entry = self._buffer.pop(0)
                self._to(entry)

    def run(self):
        while not self._stopper.is_set():
            self._fetch()
            self._stream()

            time.sleep(self.stream_rate)
        if self._fetch():  # fetch the last entries
            self._stream()

    def close(self):
        self._stopper.set()


def mongoengine_log_stream_print(entry: dict):
    if "message" in entry:
        print(entry["message"])
    else:
        raise ValueError(f"Expected 'message' in '{entry}'.")



import inspect

from mongoengine.base import BaseDocument, BaseField


class WithInstanceField(BaseField):
    def get_instance(self) -> BaseDocument:
        """
        Get the instance of the document that the field is attached to.
        :return: The instance of the document that the field is attached to.
        """

        # get stack
        stack = inspect.stack()

        # get the instance of the document that the field is attached to
        for frame in stack:
            if "self" in frame.frame.f_locals:
                if not isinstance(frame.frame.f_locals["self"], self.owner_document):
                    continue
                return frame.frame.f_locals["self"]
        raise ValueError("Could not find instance of document that the field is attached to.")



class PostInit:
    def __new__(cls, *args, **kwargs):
        # get __init__ method
        __init__ = cls.__init__

        # wrap __init__ method
        def post_init(self, *a, **kw):
            __init__(self, *a, **kw)

            # check if __post_init__ method exists
            if hasattr(self, "__post_init__"):
                self.__post_init__()

        cls.__init__ = post_init

        return super().__new__(cls)

    def __post_init__(self):
        ...



from enum import Enum

import uvicorn
from starlette.applications import Starlette
from starlette.responses import HTMLResponse
from starlette.routing import Route
from mongoengine import Document, EmbeddedDocument, StringField, IntField, FloatField, BooleanField, ListField, DictField, EmbeddedDocumentField, \
    GenericEmbeddedDocumentField

from wiederverwendbar.mongoengine import MongoengineDbSingleton
from wiederverwendbar.starlette_admin import GenericEmbeddedAdmin, GenericEmbeddedConverter, GenericEmbeddedDocumentView

# connect to database
MongoengineDbSingleton(init=True)

# Create starlette app
app = Starlette(
    routes=[
        Route(
            "/",
            lambda r: HTMLResponse('<a href="/admin/">Click me to get to Admin!</a>'),
        ),
    ],
)

# Create admin
admin = GenericEmbeddedAdmin(title="Test Admin")


class Test1(EmbeddedDocument):
    meta = {"name": "test1_qwe"}

    test_1_str = StringField()
    test_1_int = IntField()
    test_1_float = FloatField()
    test_1_bool = BooleanField()


class Test2(EmbeddedDocument):
    test_2_str = StringField()
    test_2_int = IntField()
    test_2_float = FloatField()
    test_2_bool = BooleanField()
    test_2_list = ListField(StringField())
    test_2_dict = DictField()


class TestEnum(Enum):
    A = "a"
    B = "b"
    C = "c"


class Test(Document):
    meta = {"collection": "test"}

    test_emb = EmbeddedDocumentField(Test2)
    test_gen_emb = GenericEmbeddedDocumentField(choices=[Test1, Test2], help_text="Test Generic Embedded Document Field.")
    test_gen_emb_list = ListField(GenericEmbeddedDocumentField(choices=[Test1, Test2], help_text="Test Generic Embedded Document Field."))


class TestView(GenericEmbeddedDocumentView):
    def __init__(self):
        super().__init__(document=Test, icon="fa fa-server", name="Test", label="Test", converter=GenericEmbeddedConverter())


# Add views to admin#
admin.add_view(TestView())

# Mount admin to app
admin.mount_to(app)

if __name__ == '__main__':
    uvicorn.run(app, host="0.0.0.0", port=8000)



import logging
import os
from collections.abc import Iterator
from enum import Enum
from pathlib import Path
from typing import Optional, Union

import requests

from wiederverwendbar.functions.check_hash_file import SUPPORTED_HASH_TYPES, check_hash_file

logger = logging.getLogger(__name__)


class DownloadFileState(Enum):
    START = 0
    DOWNLOAD = 1
    CHECK_HASH = 2
    END = 3


def download_file(download_url: str,
                  local_file: Union[None, str, Path] = None,
                  hash_value: Optional[str] = None,
                  hash_type: Optional[SUPPORTED_HASH_TYPES] = None,
                  method: Optional[str] = None,
                  overwrite: Optional[bool] = None,
                  chunk_size: Optional[int] = None,
                  yield_percent: Optional[bool] = None) -> Iterator[Union[DownloadFileState, int]]:
    """
    Download single file with requests module.

    :param download_url: for example https://www.example.com/test.bin
    :param local_file: file on disk (default: basename of download_url)
    :param hash_value: hash value for file
    :param hash_type: hash type for file
    :param method: GET or POST (default: GET)
    :param overwrite: overwrite local file (default: False)
    :param chunk_size: download chunk size (default: 1024)
    :param yield_percent: yield percent instead of bytes (default: True)
    :return: generator
    """

    logger.debug(f"Download file: {download_url}")

    # set default values
    if method is None:
        method = "GET"
    if overwrite is None:
        overwrite = False
    if chunk_size is None:
        chunk_size = 1024
    if yield_percent is None:
        yield_percent = True

    # parse download_url
    if local_file is None:
        local_file = os.path.basename(requests.utils.urlparse(download_url).path)
    if local_file == "":
        raise ValueError("Can't parse download_url")
    local_file = Path(local_file)

    # check if file exists
    if local_file.exists():
        if not overwrite:
            raise FileExistsError("File exists")
        logger.debug(f"Overwrite file: {local_file}")
        local_file.unlink()

    # yield start value
    logger.debug(f"Start download: {download_url} -> {local_file}")
    yield DownloadFileState.START

    response = requests.request(method=method, url=download_url, stream=True)
    total = int(response.headers.get("content-length", 0))
    bytes_downloaded = 0
    percent_downloaded = 0

    # yield start and end value
    yield 0
    if yield_percent:
        yield 100
    else:
        yield total

    # yield download value
    logger.debug(f"Download file: {download_url} -> {local_file}")
    yield DownloadFileState.DOWNLOAD

    # download file
    with open(local_file, "wb") as f:
        for data in response.iter_content(chunk_size=chunk_size):
            f.write(data)
            bytes_downloaded += len(data)
            if total > 0:
                if yield_percent:
                    percent = round(bytes_downloaded / total * 100)
                    if percent != percent_downloaded:
                        percent_downloaded = percent
                        yield percent_downloaded
                else:
                    yield bytes_downloaded

    if hash_value is not None:
        # yield check hash value
        logger.debug(f"Check hash: {download_url} -> {local_file}")
        yield DownloadFileState.CHECK_HASH

        # check hash value
        if not check_hash_file(local_file, hash_value, hash_type):
            raise ValueError("Hash value not correct")

    # yield end value
    logger.debug(f"End download: {download_url} -> {local_file}")
    yield DownloadFileState.END


def simple_download_file(download_url: str,
                         local_file: Union[None, str, Path] = None,
                         hash_value: Optional[str] = None,
                         hash_type: Optional[SUPPORTED_HASH_TYPES] = None,
                         method: Optional[str] = None,
                         overwrite: Optional[bool] = None,
                         chunk_size: Optional[int] = None,
                         raise_exception: Optional[bool] = None) -> bool:
    """
    Download single file with requests module.

    :param download_url: for example https://www.example.com/test.bin
    :param local_file: file on disk (default: basename of download_url)
    :param hash_value: hash value for file
    :param hash_type: hash type for file
    :param method: GET or POST (default: GET)
    :param overwrite: overwrite local file (default: False)
    :param chunk_size: download chunk size (default: 1024)
    :param raise_exception: raise exception on error (default: True)
    :return: True if download was successful, False otherwise
    """

    # set default values
    if raise_exception is None:
        raise_exception = True

    try:
        for state in download_file(download_url=download_url,
                                   hash_value=hash_value,
                                   hash_type=hash_type,
                                   method=method,
                                   local_file=local_file,
                                   overwrite=overwrite,
                                   chunk_size=chunk_size):
            if state == DownloadFileState.END:
                return True
    except Exception as e:
        if raise_exception:
            raise e
        logger.exception(e)

    return False



from wiederverwendbar.logger import LoggerSingleton, LoggerSettings
from wiederverwendbar.mongoengine import MongoengineDbSingleton

LoggerSingleton(name="test", settings=LoggerSettings(log_level=LoggerSettings.LogLevels.DEBUG), init=True)

if __name__ == '__main__':
    MongoengineDbSingleton(init=True)



from enum import Enum
from typing import Any

from sqlalchemy import TypeDecorator, Integer, String, Text


class _EnumValue(TypeDecorator):
    """
    Enables passing in a Python enum and storing the enum's *value* in the db.
    The default would have stored the enum's *name* (ie the string).
    """

    def __init__(self, enum_type, *args, **kwargs):
        super(_EnumValue, self).__init__(*args, **kwargs)
        self._enum_type = enum_type

    def process_bind_param(self, value, dialect):
        if isinstance(value, Enum):
            return value.value
        return value

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return self._enum_type(value)


class EnumValueStr(_EnumValue):
    impl = Text
    cache_ok = True


class EnumValueInt(_EnumValue):
    impl = Integer
    cache_ok = True


class StringBool(TypeDecorator):
    impl = Integer
    cache_ok = True

    def process_result_value(self, value, dialect):
        if value is None:
            return False
        elif value == "0":
            return False
        else:
            return True


class StringList(TypeDecorator):
    impl = String
    cache_ok = True

    def __init__(self, annotation: type, separator: str = ",", empty_is_none: bool = False, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.annotation = annotation
        self.separator = separator
        self.empty_is_none = empty_is_none

    def process_result_value(self, value, dialect):
        value_converted = []
        if value is None:
            if self.empty_is_none:
                value_converted = None
        else:
            split = value.split(self.separator)
            for item in split:
                item_converted = self.annotation(item)
                value_converted.append(item_converted)

        return value_converted



from wiederverwendbar.starlette_admin.mongoengine.auth.documents import (Session,
                                                                         User)
from wiederverwendbar.starlette_admin.mongoengine.auth.views import (AuthView,
                                                                     SessionView,
                                                                     UserView)
from wiederverwendbar.starlette_admin.mongoengine.auth.admin import MongoengineAuthAdmin
from wiederverwendbar.starlette_admin.mongoengine.auth.provider import MongoengineAdminAuthProvider
from wiederverwendbar.starlette_admin.mongoengine.auth.settings import MongoengineAdminAuthSettings



import ssl
from ipaddress import IPv4Address
from typing import Union, Optional

from pydantic import BaseModel, Field, FilePath
from uvicorn.config import SSL_PROTOCOL_VERSION


class UvicornServerSettings(BaseModel):
    server_host: Union[IPv4Address, str] = Field(default=IPv4Address("127.0.0.1"), title="Server Host", description="Server Host to bind to")
    server_port: int = Field(default=8000, title="Server Port", ge=0, le=65535, description="Server Port to bind to")
    server_reload: bool = Field(default=False, title="Server Reload", description="Server Enable auto-reload")
    server_reload_dirs: Union[None, list[str], str] = Field(default=None, title="Server Reload Dirs", description="Server Directories to watch for changes")
    server_reload_includes: Union[None, list[str], str] = Field(default=None, title="Server Reload Includes", description="File patterns to include")
    server_reload_excludes: Union[None, list[str], str] = Field(default=None, title="Server Reload Excludes", description="File patterns to exclude")
    server_reload_delay: float = Field(default=0.25, title="Server Reload Delay", ge=0.1, description="Server Delay between reloads")
    server_workers: int = Field(default=1, title="Server Workers", ge=1, le=100, description="Server Number of worker processes")
    server_ssl_keyfile: Optional[FilePath] = Field(default=None, title="Server SSL Keyfile", description="Server SSL Keyfile")
    server_ssl_certfile: Optional[FilePath] = Field(default=None, title="Server SSL Certfile", description="Server SSL Certfile")
    server_ssl_keyfile_password: Optional[str] = Field(default=None, title="Server SSL Keyfile Password", description="Server SSL Keyfile Password")
    server_ssl_version: int = Field(default=SSL_PROTOCOL_VERSION, title="Server SSL Version", description="Server SSL Version")
    server_ssl_cert_reqs: int = Field(default=ssl.CERT_NONE, title="Server SSL Cert Reqs", description="Server SSL Cert Reqs")
    server_ssl_ca_certs: Optional[FilePath] = Field(default=None, title="Server SSL CA Certs", description="Server SSL CA Certs")
    server_ssl_ciphers: str = Field(default="TLSv1", title="Server SSL Ciphers", description="Server SSL Ciphers")



from jinja2 import PackageLoader
from starlette_admin.contrib.mongoengine import Admin as MongoengineAdmin

from wiederverwendbar.starlette_admin.multi_path.admin import MultiPathAdminMeta, MultiPathAdmin
from wiederverwendbar.starlette_admin.settings.admin import SettingsAdminMeta
from wiederverwendbar.starlette_admin.form_max_fields.admin import FormMaxFieldsAdmin
from wiederverwendbar.starlette_admin.action_log.admin import ActionLogAdminMeta

# class GenericEmbeddedAdminMeta(SettingsAdminMeta, MultiPathAdminMeta): #todo: fix this
#     ...


class GenericEmbeddedAdmin(MongoengineAdmin, FormMaxFieldsAdmin, MultiPathAdmin, metaclass=ActionLogAdminMeta):
    form_max_fields = 10000
    static_files_packages = [("wiederverwendbar", "starlette_admin/mongoengine/generic_embedded_document_field/statics")]
    template_packages = [PackageLoader("wiederverwendbar", "starlette_admin/mongoengine/generic_embedded_document_field/templates")]



from mongoengine import ListField, ReferenceField, StringField

from wiederverwendbar.mongoengine import AutomaticReferenceDocument


# --- example classes ---

class Test1(AutomaticReferenceDocument):
    meta = {"collection": "test1"}

    name = StringField(required=True)
    test2_one_to_one = ReferenceField("Test2", automatic_reference="test1_one_to_one")
    test2_one_to_many = ReferenceField("Test2", automatic_reference="test1_many_to_one")
    test2_many_to_one = ListField(ReferenceField("Test2"), automatic_reference="test1_one_to_many")
    test2_many_to_many = ListField(ReferenceField("Test2"), automatic_reference="test1_many_to_many")


class Test2(AutomaticReferenceDocument):
    meta = {"collection": "test2"}

    name = StringField(required=True)
    test1_one_to_one = ReferenceField("Test1", automatic_reference="test2_one_to_one")
    test1_one_to_many = ReferenceField("Test1", automatic_reference="test2_many_to_one")
    test1_many_to_one = ListField(ReferenceField("Test1"), automatic_reference="test2_one_to_many")
    test1_many_to_many = ListField(ReferenceField("Test1"), automatic_reference="test2_many_to_many")



from jinja2 import PackageLoader
from starlette_admin.contrib.mongoengine import Admin as MongoengineAdmin

from wiederverwendbar.starlette_admin.multi_path.admin import MultiPathAdmin


class BooleanAlsoAdmin(MongoengineAdmin, MultiPathAdmin):
    static_files_packages = [("wiederverwendbar", "starlette_admin/mongoengine/boolean_also_field/statics")]
    template_packages = [PackageLoader("wiederverwendbar", "starlette_admin/mongoengine/boolean_also_field/templates")]



from jinja2 import PackageLoader

from wiederverwendbar.starlette_admin.multi_path.admin import MultiPathAdmin


class DropDownIconViewAdmin(MultiPathAdmin):
    template_packages = [PackageLoader("wiederverwendbar", "starlette_admin/drop_down_icon_view/templates")]



import logging
import os
from ipaddress import IPv4Address
from typing import Callable, Any, Union, Optional

import uvicorn
from uvicorn._types import ASGIApplication

from wiederverwendbar.uvicorn.settings import UvicornServerSettings

logger = logging.getLogger(__name__)


class UvicornServer:
    def __init__(self,
                 app: Union[ASGIApplication, Callable[..., Any], str],
                 host: Union[None, IPv4Address, str] = None,
                 port: Optional[int] = None,
                 reload: Optional[bool] = False,
                 reload_dirs: Union[list[str], str, None] = None,
                 reload_includes: Union[list[str], str, None] = None,
                 reload_excludes: Union[list[str], str, None] = None,
                 reload_delay: Optional[float] = None,
                 workers: Optional[int] = None,
                 ssl_keyfile: Optional[str] = None,
                 ssl_certfile: Union[str, os.PathLike[str], None] = None,
                 ssl_keyfile_password: Optional[str] = None,
                 ssl_version: Optional[int] = None,
                 ssl_cert_reqs: Optional[int] = None,
                 ssl_ca_certs: Optional[str] = None,
                 ssl_ciphers: Optional[str] = None,
                 auto_run: bool = True,
                 server_react_to_keyboard_interrupt: bool = True,
                 factory: bool = False,
                 settings: Optional[UvicornServerSettings] = None):
        """
        Create a new Uvicorn Server

        :param app: ASGI Application
        :param host: Host to bind to
        :param port: Port to bind to
        :param reload: Enable auto-reload
        :param reload_dirs: Reload Directories
        :param reload_includes: Reload Includes
        :param reload_excludes: Reload Excludes
        :param reload_delay: Reload Delay
        :param workers: Number of worker processes
        :param ssl_keyfile: SSL Keyfile
        :param ssl_certfile: SSL Certfile
        :param ssl_keyfile_password: SSL Keyfile Password
        :param ssl_version: SSL Version
        :param ssl_cert_reqs: SSL Cert Reqs
        :param ssl_ca_certs: SSL CA Certs
        :param ssl_ciphers: SSL Ciphers
        :param auto_run: Auto Run on creation
        :param server_react_to_keyboard_interrupt: React to Keyboard Interrupt
        :param factory: Factory
        :param settings: Uvicorn Server Settings
        """

        self.app = app
        self.settings = settings or UvicornServerSettings()

        self.host: IPv4Address = host or self.settings.server_host
        self.port: int = port or self.settings.server_port
        self.reload: bool = reload or self.settings.server_reload
        self.reload_dirs: Union[list[str], str, None] = reload_dirs or self.settings.server_reload_dirs
        self.reload_includes: Union[list[str], str, None] = reload_includes or self.settings.server_reload_includes
        self.reload_excludes: Union[list[str], str, None] = reload_excludes or self.settings.server_reload_excludes
        self.reload_delay: float = reload_delay or self.settings.server_reload_delay
        self.workers: int = workers or self.settings.server_workers
        ssl_keyfile = ssl_keyfile or self.settings.server_ssl_keyfile
        if ssl_keyfile:
            ssl_keyfile = str(ssl_keyfile)
        ssl_certfile = ssl_certfile or self.settings.server_ssl_certfile
        if ssl_certfile:
            ssl_certfile = str(ssl_certfile)
        self.ssl_keyfile: Optional[str] = ssl_keyfile
        self.ssl_certfile: Optional[str] = ssl_certfile
        self.ssl_keyfile_password: Optional[str] = ssl_keyfile_password or self.settings.server_ssl_keyfile_password
        self.ssl_version: int = ssl_version or self.settings.server_ssl_version
        self.ssl_cert_reqs: int = ssl_cert_reqs or self.settings.server_ssl_cert_reqs
        self.ssl_ca_certs: Optional[str] = str(ssl_ca_certs or self.settings.server_ssl_ca_certs)
        self.ssl_ciphers: Optional[str] = ssl_ciphers or self.settings.server_ssl_ciphers
        self.auto_run: bool = auto_run
        self.react_to_keyboard_interrupt = server_react_to_keyboard_interrupt
        self.factory: bool = factory

        logger.debug(f"Create {self}")

        if self.auto_run:
            self.run()

    def __str__(self):
        return f"{self.__class__.__name__}(host={self.host}, port={self.port})"

    def run(self) -> None:
        """
        Run the Uvicorn Server

        :return: None
        """

        logger.info(f"Run {self}")
        if not self.react_to_keyboard_interrupt:
            self._run()
        else:
            try:
                self._run()
            except KeyboardInterrupt:
                logger.info(f"Keyboard Interrupt {self}")
        logger.info(f"Stop {self}")

    def _run(self):
        uvicorn.run(self.app,
                    host=str(self.host),
                    port=self.port,
                    reload=self.reload,
                    reload_dirs=self.reload_dirs,
                    reload_includes=self.reload_includes,
                    reload_excludes=self.reload_excludes,
                    reload_delay=self.reload_delay,
                    workers=self.workers,
                    ssl_keyfile=self.ssl_keyfile,
                    ssl_certfile=self.ssl_certfile,
                    ssl_keyfile_password=self.ssl_keyfile_password,
                    ssl_version=self.ssl_version,
                    ssl_cert_reqs=self.ssl_cert_reqs,
                    ssl_ca_certs=self.ssl_ca_certs,
                    ssl_ciphers=self.ssl_ciphers,
                    factory=self.factory)



from wiederverwendbar.logger.handlers.stream_console_handler import StreamConsoleHandler

try:
    from wiederverwendbar.logger.handlers.rich_console_handler import RichConsoleHandler
except ModuleNotFoundError:
    RichConsoleHandler = None

from wiederverwendbar.logger.handlers.tar_rotating_file_handler import TarRotatingFileHandler



from datetime import datetime
from typing import Optional, Sequence

from starlette.requests import Request
from starlette.responses import Response
from starlette_admin.auth import AdminConfig, AdminUser, AuthProvider
from starlette_admin.exceptions import FormValidationError, LoginFailed
from wiederverwendbar.functions.eval import eval_value

from wiederverwendbar.starlette_admin.mongoengine.auth.documents.session import Session
from wiederverwendbar.starlette_admin.mongoengine.auth.documents.user import User
from wiederverwendbar.starlette_admin.mongoengine.auth.settings import MongoengineAdminAuthSettings
from wiederverwendbar.starlette_admin.mongoengine.helper import get_grid_fs_url


class MongoengineAdminAuthProvider(AuthProvider):
    def __init__(
            self,
            login_path: str = "/login",
            logout_path: str = "/logout",
            avatar_path: str = "/avatar",
            allow_paths: Optional[Sequence[str]] = None,
            allow_routes: Optional[Sequence[str]] = None,
            session_document_cls: Optional[type[Session]] = None,
            user_document_cls: Optional[type[User]] = None) -> None:

        # set avatar path
        self.avatar_path = avatar_path

        super().__init__(login_path=login_path, logout_path=logout_path, allow_paths=allow_paths, allow_routes=allow_routes)

        # Set default values
        self.session_document_cls = session_document_cls or Session
        self.user_document_cls = user_document_cls or User

    async def login(
            self,
            username: str,
            password: str,
            remember_me: bool,
            request: Request,
            response: Response,
    ) -> Response:
        # get settings
        settings = MongoengineAdminAuthSettings.from_state(state=request.state)

        # get user from database
        user = self.user_document_cls.objects(username=username).first()
        if user is None:
            if settings.admin_debug:
                raise FormValidationError({"username": "User not found!"})
            raise LoginFailed("Invalid username or password!")

        # check if user password is set
        if user.password is None:
            if not settings.admin_user_allows_empty_password_login:
                if settings.admin_debug:
                    raise FormValidationError({"password": "Password is not set for this user!"})
                raise LoginFailed("Invalid username or password!")
        else:
            # validate password
            if not user.password.verify_password(password):
                if settings.admin_debug:
                    raise FormValidationError({"password": "Invalid password!"})
                raise LoginFailed("Invalid username or password!")

        # check if user have already a session
        if settings.admin_session_only_one:
            session = self.session_document_cls.objects(user=user).first()
            if session is not None:
                session.delete()

        # create new session
        session: Session = user.create_session_from_request(request=request)

        # save session id in session
        request.session.update({"session_id": str(session.id)})

        return response

    async def is_authenticated(self, request) -> bool:
        # get session
        session = self.session_document_cls.get_session_from_request(request)
        if session is None:
            return False

        # update user-agent
        session.user_agent = request.headers.get("User-Agent", "")

        # update last access
        session.last_access = datetime.now()

        # save session
        session.save()

        # save session in request state
        request.state.session = session

        return True

    def get_admin_config(self, request: Request) -> AdminConfig:
        # get settings
        settings = MongoengineAdminAuthSettings.from_state(state=request.state)

        # get session
        if not hasattr(request.state, "session"):
            return AdminConfig(app_title=settings.admin_title, logo_url=settings.admin_logo_url)
        session = request.state.session

        # Update app title according to current_user
        custom_app_title = eval_value("Admin app title", settings.admin_user_app_title, session=session)

        # Update logo url according to current_user
        custom_logo_url = settings.admin_logo_url
        if session.user.company_logo is not None:
            custom_logo_url = str(request.url_for(settings.admin_route_name,
                                                  path=str(settings.admin_static_company_logo_dir / session.user.company_logo).replace("\\", "/")))

        return AdminConfig(app_title=custom_app_title, logo_url=custom_logo_url)

    def get_admin_user(self, request: Request) -> AdminUser:
        # get session
        if not hasattr(request.state, "session"):
            return AdminUser(username="")
        session = request.state.session

        # get avatar url
        avatar_url = get_grid_fs_url(session.user.avatar, request=request)

        return AdminUser(username=session.user.username, photo_url=avatar_url)

    async def logout(self, request: Request, response: Response) -> Response:
        #
        session = self.session_document_cls.get_session_from_request(request)
        if session is None:
            raise LoginFailed("Invalid session!")

        # delete session
        session.delete()

        # clear session
        request.session.clear()

        return response



from wiederverwendbar.sqlalchemy.base import (Base,
                                              DictColumn,
                                              DeserializationError,
                                              MappingError,
                                              SelectionError,
                                              SerializationError)
from wiederverwendbar.sqlalchemy.db import (SqlalchemyDb)
from wiederverwendbar.sqlalchemy.raise_has_not_attr import (raise_has_not_attr)
from wiederverwendbar.sqlalchemy.settings import (SqlalchemySettings)
from wiederverwendbar.sqlalchemy.singleton import (SqlalchemyDbSingleton)
from wiederverwendbar.sqlalchemy.types import (EnumValueInt,
                                               EnumValueStr,
                                               StringBool,
                                               StringList)



import logging
import warnings
from typing import Optional, Sequence, Tuple, Any, Union

from starlette.middleware import Middleware
from starlette.middleware.sessions import SessionMiddleware
from starlette.requests import Request

from starlette_admin.contrib.mongoengine import Admin as MongoengineAdmin
from starlette_admin.i18n import I18nConfig
from starlette_admin.views import CustomView
from starlette_admin.auth import BaseAuthProvider

from wiederverwendbar.starlette_admin.drop_down_icon_view.admin import DropDownIconViewAdmin
from wiederverwendbar.starlette_admin.mongoengine.auth.settings import MongoengineAdminAuthSettings
from wiederverwendbar.starlette_admin.mongoengine.auth.views.auth import AuthView
from wiederverwendbar.starlette_admin.mongoengine.auth.provider import MongoengineAdminAuthProvider
from wiederverwendbar.starlette_admin.mongoengine.auth.documents.session import Session
from wiederverwendbar.starlette_admin.mongoengine.auth.views.session import SessionView
from wiederverwendbar.starlette_admin.mongoengine.auth.views.user import UserView
from wiederverwendbar.starlette_admin.mongoengine.auth.documents.user import User
from wiederverwendbar.starlette_admin.multi_path.admin import MultiPathAdminMeta
from wiederverwendbar.starlette_admin.settings.admin import SettingsAdminMeta, SettingsAdmin

logger = logging.getLogger(__name__)

class MongoengineAuthAdminMeta(SettingsAdminMeta, MultiPathAdminMeta):
    ...


class MongoengineAuthAdmin(MongoengineAdmin, SettingsAdmin, DropDownIconViewAdmin, metaclass=MongoengineAuthAdminMeta):
    settings_class = MongoengineAdminAuthSettings

    def __init__(
            self,
            title: Optional[str] = None,
            base_url: Optional[str] = None,
            route_name: Optional[str] = None,
            logo_url: Optional[str] = None,
            login_logo_url: Optional[str] = None,
            templates_dir: Optional[str] = None,
            statics_dir: Optional[str] = None,
            index_view: Optional[CustomView] = None,
            auth_view: Union[None, AuthView, bool] = None,
            user_document: Optional[type[User]] = None,
            user_view: Optional[UserView] = None,
            session_document: Optional[type[Session]] = None,
            session_view: Optional[SessionView] = None,
            auth_provider: Optional[BaseAuthProvider] = None,
            middlewares: Optional[Sequence[Middleware]] = None,
            session_middleware: Optional[type[SessionMiddleware]] = None,
            debug: Optional[bool] = None,
            i18n_config: Optional[I18nConfig] = None,
            favicon_url: Optional[str] = None,
            settings: Optional[MongoengineAdminAuthSettings] = None
    ):
        # set documents
        self.user_document = user_document or User
        self.session_document = session_document or Session

        # set views
        if auth_view is None:
            auth_view = AuthView()
        self.auth_view = auth_view
        self.user_view = user_view or UserView(document=self.user_document, company_logo_choices_loader=self.user_company_logo_files_loader)
        self.session_view = session_view or SessionView(document=self.session_document)

        # set auth_provider
        if settings.admin_auth:
            auth_provider = auth_provider or MongoengineAdminAuthProvider(login_path=settings.admin_login_path,
                                                                          logout_path=settings.admin_logout_path,
                                                                          avatar_path=f"/{self.user_view.identity}/avatar",
                                                                          allow_routes=settings.admin_allow_routes,
                                                                          session_document_cls=session_document,
                                                                          user_document_cls=user_document)
        else:
            auth_provider = None
        self.auth_provider = auth_provider

        super().__init__(
            title=title,
            base_url=base_url,
            route_name=route_name,
            logo_url=logo_url,
            login_logo_url=login_logo_url,
            templates_dir=templates_dir,
            statics_dir=statics_dir,
            index_view=index_view,
            auth_provider=auth_provider,
            middlewares=middlewares,
            session_middleware=session_middleware,
            debug=debug,
            i18n_config=i18n_config,
            favicon_url=favicon_url,
            settings=settings,
        )

        # create views
        if self.auth_view:
            self.auth_view.views = [self.user_view, self.session_view]
            self.add_view(self.auth_view)
        else:
            self.add_view(self.user_view)
            self.add_view(self.session_view)

        # check if superuser is set
        if settings.admin_superuser_username is not None:
            # check if superuser exists
            if not self.user_document.objects(username=settings.admin_superuser_username).first():
                if settings.admin_superuser_auto_create:
                    # create superuser
                    logger.info(f"Creating superuser with username '{settings.admin_superuser_username}' and password '{settings.admin_superuser_username}'")
                    superuser = self.user_document(username=settings.admin_superuser_username)
                    superuser.password = settings.admin_superuser_username
                    superuser.save()
                else:
                    warnings.warn(f"Superuser with username '{settings.admin_superuser_username}' does not exist!", UserWarning)

    def user_company_logo_files_loader(self, request: Request) -> Sequence[Tuple[Any, str]]:
        if not self.settings.admin_static_company_logo_dir:
            return []
        company_logo_files = []
        for file in self.settings.admin_static_company_logo_dir.iterdir():
            if not file.is_file():
                continue
            if file.suffix not in self.settings.admin_company_logos_suffixes:
                continue
            company_logo_files.append((file.name, file.name))
        return company_logo_files



from wiederverwendbar.threading.extended_thread import (ThreadInterrupt,
                                                        ThreadLoopContinue,
                                                        ThreadStop,
                                                        ThreadKill,
                                                        ThreadWatchdogError,
                                                        ExtendedThread,
                                                        handle_exception)



from wiederverwendbar.starlette_admin.mongoengine.auth.views.auth import AuthView
from wiederverwendbar.starlette_admin.mongoengine.auth.views.session import SessionView
from wiederverwendbar.starlette_admin.mongoengine.auth.views.user import UserView



from typing import Type, Optional

from starlette.requests import Request
from starlette_admin.contrib.mongoengine.converters import BaseMongoEngineModelConverter

from wiederverwendbar.starlette_admin.mongoengine.auth.documents.session import Session
from wiederverwendbar.starlette_admin.mongoengine.view import MongoengineModelView


class SessionView(MongoengineModelView):
    exclude_fields_from_list = [Session.id]
    exclude_fields_from_detail = [Session.id]
    exclude_fields_from_create = [Session.id, Session.user, Session.app_name, Session.user_agent, Session.created, Session.last_access]
    exclude_fields_from_edit = [Session.id, Session.user, Session.app_name, Session.user_agent, Session.created, Session.last_access]

    def __init__(
            self,
            document: Type[Session],
            icon: Optional[str] = None,
            name: Optional[str] = None,
            label: Optional[str] = None,
            identity: Optional[str] = None,
            converter: Optional[BaseMongoEngineModelConverter] = None,
    ):
        # set default values
        document = document or Session
        icon = icon or "fa-solid fa-list"
        name = name or "Session"
        label = label or "Sitzung"

        super().__init__(document=document,
                         icon=icon,
                         name=name,
                         label=label,
                         identity=identity,
                         converter=converter)
        for field in self.fields:
            if field.name == "user":
                field.label = "Benutzer"
            elif field.name == "created":
                field.label = "Erstellt"
            elif field.name == "last_access":
                field.label = "Letzter Zugriff"

    def can_create(self, request: Request) -> bool:
        return False

    def can_edit(self, request: Request) -> bool:
        return False



import logging as _logging
import inspect as _inspect

import threading as _threading
import time as _time
from itertools import count as _count
from typing import Any as _Any, Optional as _Optional, Union as _Union
from datetime import datetime as _datetime, timedelta as _timedelta
from enum import Enum as _Enum

from bson import ObjectId as _ObjectId
from mongoengine import DoesNotExist as _DoesNotExist, ValidationError as _ValidationError, Document as _Document, EmbeddedDocument as _EmbeddedDocument, EnumField as _EnumField, \
    DateTimeField as _DateTimeField, \
    DictField as _DictField, StringField as _StringField, ReferenceField as _ReferenceField, FloatField as _FloatField, IntField as _IntField, \
    EmbeddedDocumentListField as _EmbeddedDocumentListField

from wiederverwendbar.functions.datetime import local_now as _local_now, to_local as _to_local
from wiederverwendbar.logger.helper import remove_logger as _remove_logger
from wiederverwendbar.logger.context import LoggingContext as _LoggingContext
from wiederverwendbar.logger.singleton import SubLogger as _SubLogger
from wiederverwendbar.mongoengine.logger.streamer import MongoengineLogStreamer as _MongoengineLogStreamer
from wiederverwendbar.mongoengine.logger.documets import MongoengineLogDocument as _MongoengineLogDocument
from wiederverwendbar.mongoengine.logger.handlers import MongoengineLogHandler as _MongoengineLogHandler
from wiederverwendbar.threading import ExtendedThread as _ExtendedThread, handle_exception as _handle_exception, ThreadLoopContinue as _ThreadLoopContinue, \
    ThreadStop as _ThreadStop

MODULE_NAME = "task_manager"
DEFAULT_LOG_LEVEL = _logging.INFO
MANAGER_NAMESPACE_NAME = f"{MODULE_NAME}.manager"
WORKER_NAMESPACE_NAME = f"{MODULE_NAME}.worker"
TASK_NAMESPACE_NAME = f"{MODULE_NAME}.task"


class WorkerSignalType(_Enum):
    STOP = "stop"
    CANCEL = "cancel"


class _WorkerSignal(_EmbeddedDocument):
    type: WorkerSignalType = _EnumField(WorkerSignalType, required=True)
    send_at: _datetime = _DateTimeField(required=True)
    content: dict[_Any, _Any] = _DictField()


class WorkerState(_Enum):
    BUSY = "busy"
    IDLE = "idle"
    QUIT = "quit"


class _WorkerDocument(_Document):
    meta = {"collection": WORKER_NAMESPACE_NAME}

    name: str = _StringField(required=True, unique_with="manager")
    manager: str = _StringField(required=True)
    state: WorkerState = _EnumField(WorkerState, required=True)
    started_at: _Optional[_datetime] = _DateTimeField()
    last_seen: _Optional[_datetime] = _DateTimeField()
    delay: _Optional[float] = _FloatField()
    current_task: _Optional["_TaskDocument"] = _ReferenceField("_TaskDocument")
    signals: list[_WorkerSignal] = _EmbeddedDocumentListField(_WorkerSignal)


class _WorkerLogDocument(_MongoengineLogDocument):
    meta = {"collection": f"{WORKER_NAMESPACE_NAME}.log",
            "indexes": ["owner"]}
    owner: _WorkerDocument = _ReferenceField(_WorkerDocument, required=True)


class TaskState(_Enum):
    NEW = "new"
    RUNNING = "running"  # running states
    CANCELED = "canceled"  # final states
    FINISHED = "finished"  # final states
    FAILED = "failed"  # final states

    @classmethod
    def waiting_states(cls) -> list["TaskState"]:
        return [cls.NEW]

    @classmethod
    def running_states(cls) -> list["TaskState"]:
        return [cls.RUNNING]

    @classmethod
    def failed_states(cls) -> list["TaskState"]:
        return [cls.CANCELED, cls.FAILED]

    @classmethod
    def final_states(cls) -> list["TaskState"]:
        return [cls.FINISHED] + cls.failed_states()


class _TaskCancelSignal(_ThreadLoopContinue):
    ...


class _TaskDocument(_Document):
    meta = {"collection": TASK_NAMESPACE_NAME}

    name: str = _StringField(required=True)
    manager: str = _StringField(required=True)
    state: TaskState = _EnumField(TaskState, required=True)
    worker: _Optional[_WorkerDocument] = _ReferenceField(_WorkerDocument)
    log_level: int = _IntField(required=True)
    created_at: _datetime = _DateTimeField(required=True)
    due_at: _datetime = _DateTimeField(required=True)
    started_at: _Optional[_datetime] = _DateTimeField()
    ended_at: _Optional[_datetime] = _DateTimeField()
    params: dict[str, _Any] = _DictField(required=True)
    result: _Optional[dict[str, _Any]] = _DictField()


class _TaskLogDocument(_MongoengineLogDocument):
    meta = {"collection": f"{TASK_NAMESPACE_NAME}.log",
            "indexes": ["owner"]}
    owner: _TaskDocument = _ReferenceField(_TaskDocument, required=True)


class _BaseProxy:
    proxy_document_cls: _Union[type[_WorkerDocument], type[_TaskDocument]] = None
    log_document_cls: _Union[type[_WorkerLogDocument], type[_WorkerLogDocument]] = None

    def __init__(self, object_id: _ObjectId = None):
        self._proxy_document: _Union[_WorkerDocument, _TaskDocument] = self.proxy_document_cls.objects.get(id=object_id)

    def __str__(self):
        return f"{self.__class__.__name__}(name={self.name}, manager={self.manager}, state={self.state})"

    @property
    def id(self) -> _ObjectId:
        return self._proxy_document.id

    @property
    def name(self) -> str:
        return self._proxy_document.name

    @property
    def manager(self) -> "Manager":
        return Manager.get_manager(self._proxy_document.manager)

    @property
    def state(self) -> _Union[WorkerState, TaskState]:
        self.reload()
        return self._proxy_document.state

    @property
    def started_at(self) -> _Optional[_datetime]:
        self.reload()
        return self._proxy_document.started_at

    def reload(self) -> None:
        self._proxy_document.reload()

    def log_streamer(self,
                     to: _Optional[callable] = None,
                     begin: _Optional[_datetime] = None,
                     log_stream_rate: _Optional[float] = None) -> _MongoengineLogStreamer:
        if begin is None:
            begin = self.started_at
        if log_stream_rate is None:
            log_stream_rate = self.manager.log_stream_rate
        return _MongoengineLogStreamer(log_document=self.log_document_cls,
                                       search={"owner": self._proxy_document},
                                       to=to,
                                       name=f"{self.name}.log_streamer",
                                       begin=begin,
                                       stream_rate=log_stream_rate)

    def wait_for_state(self, *states: _Union[WorkerState, TaskState], timeout: _Optional[float] = None) -> None:
        start_time = _time.perf_counter()
        while self.state not in states:
            if timeout is not None and _time.perf_counter() - start_time > timeout:
                raise TimeoutError(f"Timeout while waiting for task '{self.name}' to reach state '{states}'")
            _time.sleep(0.001)


class Worker(_BaseProxy):
    proxy_document_cls = _WorkerDocument
    log_document_cls = _WorkerLogDocument

    def __init__(self, object_id: _ObjectId = None):
        if object_id is None:
            stack = _inspect.stack()
            for frame_info in stack[1:]:
                for f_local_name, f_local in dict(frame_info.frame.f_locals).items():
                    if isinstance(f_local, _WorkerThread):
                        object_id = f_local.worker_document.id
                        break
                if object_id is not None:
                    break
            if object_id is None:
                raise ValueError("No worker found in stack")
        super().__init__(object_id=object_id)

    @property
    def state(self) -> WorkerState:
        return super().state

    @property
    def last_seen(self) -> _datetime:
        self.reload()
        return self._proxy_document.last_seen

    @property
    def delay(self) -> float:
        self.reload()
        return self._proxy_document.delay

    @property
    def current_task(self) -> _Optional["ScheduledTask"]:
        self.reload()
        if self._proxy_document.current_task is None:
            return None
        return ScheduledTask(object_id=self._proxy_document.current_task.id)

    def wait_for_state(self, *states: WorkerState, timeout: _Optional[float] = None) -> None:
        super().wait_for_state(*states, timeout=timeout)

    def wait_for_task(self, timeout: _Optional[float] = None) -> None:
        current_task = self.current_task
        if current_task is None:
            return
        current_task.wait_for_end(timeout=timeout)

    # def terminate(self, wait: bool = False, timeout: _Optional[float] = None) -> None:
    #     if self.state == WorkerState.TERMINATE:
    #         return
    #     elif self.state == WorkerState.QUIT:
    #         raise ValueError(f"Worker '{self.name}' is already quitting")
    #     self._proxy_document.state = WorkerState.TERMINATE
    #     self._proxy_document.save()
    #     if wait:
    #         self.wait_for_state(WorkerState.TERMINATE, timeout=timeout)


class ScheduledTask(_BaseProxy):
    proxy_document_cls = _TaskDocument
    log_document_cls = _TaskLogDocument

    def __init__(self, object_id: _ObjectId = None):
        if object_id is None:
            stack = _inspect.stack()
            for frame_info in stack[1:]:
                for f_local_name, f_local in dict(frame_info.frame.f_locals).items():
                    if isinstance(f_local, _WorkerThread):
                        object_id = f_local.current_task.id
                        break
                if object_id is not None:
                    break
            if object_id is None:
                raise ValueError("No task found in stack")
        super().__init__(object_id=object_id)

    @property
    def state(self) -> TaskState:
        return super().state

    @property
    def worker(self) -> _Optional[Worker]:
        self.reload()
        if self._proxy_document.worker is None:
            return None
        return Worker(object_id=self._proxy_document.worker.id)

    @property
    def created_at(self) -> _datetime:
        return self._proxy_document.created_at

    @property
    def due_at(self) -> _datetime:
        return self._proxy_document.due_at

    @property
    def ended_at(self) -> _Optional[_datetime]:
        self.reload()
        return self._proxy_document.ended_at

    @property
    def params(self) -> dict[str, _Any]:
        return self._proxy_document.params.copy()

    @property
    def result(self) -> _Optional[dict[str, _Any]]:
        self.reload()
        if self._proxy_document.result is None:
            return None
        return self._proxy_document.result.copy()

    @property
    def duration(self) -> _Optional[float]:
        self.reload()
        if self._proxy_document.started_at is None or self._proxy_document.ended_at is None:
            return None
        return (self._proxy_document.ended_at - self._proxy_document.started_at).total_seconds()

    @property
    def is_running(self) -> bool:
        return self.state in TaskState.running_states()

    def wait_for_state(self, *states: TaskState, timeout: _Optional[float] = None) -> None:
        super().wait_for_state(*states, timeout=timeout)

    def wait_for_end(self, timeout: _Optional[float] = None) -> None:
        self.wait_for_state(*TaskState.final_states(), timeout=timeout)

    # def cancel(self, wait: bool = False, timeout: _Optional[float] = None) -> None:
    #     self.reload()
    #     if self._proxy_document.state == TaskState.CANCELING:
    #         return
    #     if self._proxy_document.state == TaskState.CANCELED:
    #         raise ValueError(f"Task '{self.name}' is already canceled")
    #     elif self._proxy_document.state == TaskState.FINISHED:
    #         raise ValueError(f"Task '{self.name}' is already finished")
    #     elif self._proxy_document.state == TaskState.FAILED:
    #         raise ValueError(f"Task '{self.name}' is already failed")
    #
    #     self._proxy_document.state = TaskState.CANCELING
    #     self._proxy_document.save()
    #
    #     if wait:
    #         self.wait_for_state(TaskState.CANCELED, timeout=timeout)


_worker_counter = _count().__next__
_worker_counter()  # skip 0


class _WorkerThread(_ExtendedThread):
    def __init__(self,
                 name: str,
                 manager: "Manager",
                 document: _WorkerDocument,
                 log_level: int,
                 log_push_rate: _Optional[float] = None,
                 log_push_max_entries: _Optional[int] = None,
                 task_ignore_loggers_equal: _Optional[list[str]] = None,
                 task_ignore_loggers_like: _Optional[list[str]] = None,
                 logger: _Optional[_logging.Logger] = None,
                 loop_sleep_time: _Optional[float] = None):

        self._manager: "Manager" = manager

        # define document attributes
        self._state: WorkerState = WorkerState.BUSY
        self._last_seen: _datetime = _local_now()
        self._current_task: _Optional[_TaskDocument] = None

        # define worker attributes
        self._log_level: int = log_level
        self._logger_handler: _Optional[_MongoengineLogHandler] = None
        self._worker_document: _WorkerDocument = document
        self._log_push_rate: _Optional[float] = log_push_rate
        self._log_push_max_entries: _Optional[int] = log_push_max_entries
        self._task_logger: _Optional[_logging.Logger] = None
        self._task_logger_handler: _Optional[_MongoengineLogHandler] = None
        if task_ignore_loggers_equal is None:
            task_ignore_loggers_equal = []
        self._task_ignore_loggers_equal: list[str] = task_ignore_loggers_equal
        if task_ignore_loggers_like is None:
            task_ignore_loggers_like = []
        self._task_ignore_loggers_like: list[str] = task_ignore_loggers_like
        self._task_result: _Optional[dict[str, _Any]] = None
        self._task_state: _Optional[TaskState] = None

        super().__init__(name=name,
                         cls_name=name,
                         logger=logger,
                         loop_sleep_time=loop_sleep_time,
                         loop_stop_on_other_exception=False,
                         watchdog_target=self._watchdog_target)

    # --- threads ---

    @classmethod
    def _watchdog_target(cls, thread) -> bool:
        thread: _WorkerThread

        # check if manager thread is alive
        if not thread.manager_thread.is_alive():
            thread.stop()
            return False

        # check if worker signal
        thread.worker_document.reload()
        for signal in thread.worker_document.signals:
            print(signal)

        return True

    # --- properties ---

    @property
    def worker_document(self) -> _WorkerDocument:
        with self.lock:
            return self._worker_document

    @property
    def state(self) -> WorkerState:
        with self.lock:
            return self._state

    @state.setter
    def state(self, value: WorkerState) -> None:
        with self.lock:
            if value == self._state:
                return
            self._logger.debug(f"Worker state changed from '{self._state}' to '{value}'")
            self._state = value
            self._worker_document.state = self._state
            self._worker_document.save()

    @property
    def last_seen(self) -> _datetime:
        with self.lock:
            return self._last_seen

    @last_seen.setter
    def last_seen(self, value: _datetime) -> None:
        with self.lock:
            self._last_seen = value
            self._worker_document.last_seen = self._last_seen
            self._worker_document.save()

    @property
    def current_task(self) -> _Optional[_TaskDocument]:
        with self.lock:
            return self._current_task

    @current_task.setter
    def current_task(self, value: _Optional[_TaskDocument]) -> None:
        with self.lock:
            self._current_task = value
            self._worker_document.current_task = self._current_task
            self._worker_document.save()

    @property
    def log_push_rate(self) -> float:
        with self.lock:
            if self._log_push_rate is None:
                return self._manager.log_push_rate
            return self._log_push_rate

    @log_push_rate.setter
    def log_push_rate(self, value: float) -> None:
        with self.lock:
            self._log_push_rate = value

    @property
    def log_push_max_entries(self) -> int:
        with self.lock:
            if self._log_push_max_entries is None:
                return self._manager.log_push_max_entries
            return self._log_push_max_entries

    @log_push_max_entries.setter
    def log_push_max_entries(self, value: int) -> None:
        with self.lock:
            self._log_push_max_entries = value

    @property
    def loop_sleep_time(self) -> _Optional[float]:
        with self.lock:
            if self._loop_sleep_time is None:
                return self._manager.worker_loop_sleep_time
        return super().loop_sleep_time

    @property
    def manager_thread(self):
        return self._manager.thread

    @property
    def task_ignore_loggers_equal(self) -> list[str]:
        task_ignore_loggers_equal = self._manager.task_ignore_loggers_equal.copy()
        with self.lock:
            task_ignore_loggers_equal.extend(self._task_ignore_loggers_equal)
            return task_ignore_loggers_equal

    @task_ignore_loggers_equal.setter
    def task_ignore_loggers_equal(self, value: list[str]):
        with self.lock:
            self._task_ignore_loggers_equal = value

    @property
    def task_ignore_loggers_like(self) -> list[str]:
        task_ignore_loggers_like = self._manager.task_ignore_loggers_like.copy()
        with self.lock:
            task_ignore_loggers_like.extend(self._task_ignore_loggers_like)
            if "pymongo" not in task_ignore_loggers_like:
                task_ignore_loggers_like.append("pymongo")
            return task_ignore_loggers_like

    @task_ignore_loggers_like.setter
    def task_ignore_loggers_like(self, value: list[str]):
        with self.lock:
            self._task_ignore_loggers_like = value

    # --- methods ---

    def on_start(self) -> None:
        # create worker logger handler
        self._logger_handler = _MongoengineLogHandler(document=_WorkerLogDocument,
                                                      document_kwargs={"owner": self.worker_document},
                                                      buffer_size=self._manager.log_push_max_entries,
                                                      buffer_periodical_flush_timing=self._manager.log_push_rate)

        # configure logger
        if not isinstance(self._logger, _SubLogger):
            self._logger.setLevel(self._log_level)
            self._logger.addHandler(self._logger_handler)
        else:
            with self._logger.reconfigure():
                self._logger.setLevel(self._log_level)
                self._logger.addHandler(self._logger_handler)

        # set all running tasks for this worker to failed

        for running_task in _TaskDocument.objects(worker=self.worker_document,
                                                  state__in=TaskState.running_states()):
            self.logger.debug(f"Setting task '{running_task.name}' to state '{TaskState.CANCELED}'")

            # set task attributes
            running_task.state = TaskState.CANCELED
            running_task.ended_at = _local_now()
            running_task.result = {"error": "Worker restarted"}

            # save task
            running_task.save()
        with self.lock:
            self._started_at = _local_now()
            self._worker_document.started_at = self._started_at
            self._worker_document.save()

    def on_loop_start(self) -> None:
        # set state to idle
        self.state = WorkerState.IDLE

        with self.lock:
            # set last seen and delay
            self._worker_document.last_seen = _local_now()
            self._worker_document.delay = self._loop_delay
            self._worker_document.save()

    def loop(self) -> None:
        # get next due task
        self.current_task: _TaskDocument = _TaskDocument.objects(manager=self._manager.name,
                                                                 due_at__lte=_local_now(),
                                                                 state=TaskState.NEW,
                                                                 worker=None).order_by("due_at").first()
        if self.current_task is not None:
            self.on_task_start()

            # run task
            raise_at_end = None
            try:
                self.task()
            except (_ThreadLoopContinue, _ThreadStop) as e:
                if isinstance(e, _ThreadStop):
                    self._task_result = {"error": "Worker stopped"}
                elif isinstance(e, _ThreadLoopContinue):
                    self._task_result = {"error": "Worker continued"}
                self._task_state = TaskState.CANCELED
                raise_at_end = e
            except Exception as e:
                self._task_result = {"error": str(e)}
                self._task_state = TaskState.FAILED
                _handle_exception(msg=f"Error while running task '{self.current_task.name}'", e=e, logger=self._logger)

            self.on_task_end()

            # set current task to None
            self.current_task = None

            if raise_at_end is not None:
                raise raise_at_end

    def on_task_start(self) -> None:
        # start task
        self.current_task.state = TaskState.RUNNING
        self.current_task.worker = self.worker_document
        self.current_task.started_at = _local_now()
        self.current_task.save()

        # set state to busy
        self.state = WorkerState.BUSY

        self._logger.info(f"Running task '{self.current_task.name}'")

        # create logger
        self._task_logger = _logging.getLogger(f"{TASK_NAMESPACE_NAME}.{self.current_task.name}")

        # add logger handler
        self._task_logger_handler = _MongoengineLogHandler(document=_TaskLogDocument,
                                                           document_kwargs={"owner": self.current_task},
                                                           buffer_size=self.log_push_max_entries,
                                                           buffer_periodical_flush_timing=self.log_push_rate)
        if not isinstance(self._task_logger, _SubLogger):
            self._task_logger.setLevel(self.current_task.log_level)
            self._task_logger.addHandler(self._task_logger_handler)
        else:
            with self._task_logger.reconfigure():
                self._task_logger.setLevel(self.current_task.log_level)
                self._task_logger.addHandler(self._task_logger_handler)

        self._task_result = None
        self._task_state = None

    def task(self):
        with _LoggingContext(context_logger=self._task_logger,
                             ignore_loggers_equal=self.task_ignore_loggers_equal,
                             ignore_loggers_like=self.task_ignore_loggers_like):
            task_func = self._manager.get_task_func(self.current_task.name)
            self._task_result = task_func(**self.current_task.params)
            self._task_state = TaskState.FINISHED

    def on_task_end(self) -> None:
        # close logger handler and remove it
        self._task_logger_handler.close()
        if not isinstance(self._task_logger, _SubLogger):
            self._task_logger.removeHandler(self._task_logger_handler)
        else:
            with self._task_logger.reconfigure():
                self._task_logger.removeHandler(self._task_logger_handler)

        # remove logger
        _remove_logger(self._task_logger)

        self._logger.info(f"Finishing task '{self.current_task.name}'")

        try:
            # set task attributes
            self.current_task.state = self._task_state
            self.current_task.ended_at = _local_now()
            self.current_task.result = self._task_result

            # save task
            self.current_task.save()
        except _ValidationError as e:
            self.current_task.state = TaskState.FAILED
            self.current_task.result = {"error": _handle_exception(msg=f"Validation error while saving task '{self.current_task.name}'", e=e, logger=self._logger)}
            self.current_task.save()

    def on_loop_end(self) -> None:
        ...

    def on_stop(self) -> None:
        # set state to quit
        self.state = WorkerState.QUIT

    def on_end(self) -> None:
        # close logger handler and remove it
        self._logger_handler.close()
        if not isinstance(self._logger, _SubLogger):
            self._logger.removeHandler(self._logger_handler)
        else:
            with self._logger.reconfigure():
                self._logger.removeHandler(self._logger_handler)

        # remove logger
        _remove_logger(self._logger)


class Manager:
    _manager_lock = _threading.Lock()
    managers: dict[str, "Manager"] = {}

    def __init__(self,
                 name: _Optional[str] = None,
                 log_level: _Optional[int] = None,
                 log_stream_rate: _Optional[float] = None,
                 log_push_rate: _Optional[float] = None,
                 log_push_max_entries: _Optional[int] = None,
                 minimum_last_seen_time_for_other_worker: _Optional[int] = None,
                 worker_loop_sleep_time: _Optional[float] = None,
                 task_ignore_loggers_equal: _Optional[list[str]] = None,
                 task_ignore_loggers_like: _Optional[list[str]] = None, ):
        self._lock = _threading.Lock()

        if name is None:
            name = "default"
        self._name: str = name

        self._logger = _logging.getLogger(f"{MANAGER_NAMESPACE_NAME}.{self.name}")
        if log_level is None:
            if self._logger.level == _logging.NOTSET:
                log_level = DEFAULT_LOG_LEVEL
            else:
                log_level = self._logger.level
        if not isinstance(self._logger, _SubLogger):
            self._logger.setLevel(log_level)
        else:
            with self._logger.reconfigure():
                self._logger.setLevel(log_level)

        if log_stream_rate is None:
            log_stream_rate = 0.001
        self._log_stream_rate: float = log_stream_rate

        if log_push_rate is None:
            log_push_rate = 1.0
        self._log_push_rate: float = log_push_rate

        if log_push_max_entries is None:
            log_push_max_entries = 10
        self._log_push_max_entries: int = log_push_max_entries

        if minimum_last_seen_time_for_other_worker is None:
            minimum_last_seen_time_for_other_worker = 10
        self._minimum_last_seen_time_for_other_worker: int = minimum_last_seen_time_for_other_worker

        self._workers: dict[str, _ObjectId] = {}

        if worker_loop_sleep_time is None:
            worker_loop_sleep_time = 1
        self._worker_loop_sleep_time: float = worker_loop_sleep_time

        if task_ignore_loggers_equal is None:
            task_ignore_loggers_equal = []
        self._task_ignore_loggers_equal: list[str] = task_ignore_loggers_equal
        if task_ignore_loggers_like is None:
            task_ignore_loggers_like = []
        self._task_ignore_loggers_like: list[str] = task_ignore_loggers_like

        self._thread = _threading.current_thread()

        self._registered_tasks: dict[str, callable] = {}

        self._registered_tasks_param: dict[str, dict[str, type]] = {}

        self._registered_tasks_param_defaults: dict[str, dict[str, _Any]] = {}

        # add manager to managers
        with self._manager_lock:
            if self.name in self.managers:
                raise ValueError(f"Manager with name '{self.name}' already exists")
            self.managers[self.name] = self

        self._logger.info("Manager created")

    @classmethod
    def get_manager(cls, name: str) -> "Manager":
        with cls._manager_lock:
            if name not in cls.managers:
                raise ValueError(f"No manager with name '{name}' found")
            return cls.managers[name]

    @property
    def name(self) -> str:
        with self._lock:
            return self._name

    @property
    def log_level(self) -> int:
        return self._logger.level

    @log_level.setter
    def log_level(self, value: int) -> None:
        self._logger.setLevel(value)

    @property
    def log_stream_rate(self) -> float:
        with self._lock:
            return self._log_stream_rate

    @log_stream_rate.setter
    def log_stream_rate(self, value: float) -> None:
        with self._lock:
            self._log_stream_rate = value

    @property
    def log_push_rate(self) -> float:
        with self._lock:
            return self._log_push_rate

    @log_push_rate.setter
    def log_push_rate(self, value: float) -> None:
        with self._lock:
            self._log_push_rate = value

    @property
    def log_push_max_entries(self) -> int:
        with self._lock:
            return self._log_push_max_entries

    @log_push_max_entries.setter
    def log_push_max_entries(self, value: int) -> None:
        with self._lock:
            self._log_push_max_entries = value

    @property
    def thread(self) -> _threading.Thread:
        with self._lock:
            return self._thread

    # --- worker management ---

    @property
    def minimum_last_seen_time_for_other_worker(self) -> _Optional[int]:
        with self._lock:
            return self._minimum_last_seen_time_for_other_worker

    @minimum_last_seen_time_for_other_worker.setter
    def minimum_last_seen_time_for_other_worker(self, value: _Optional[int]) -> None:
        with self._lock:
            self._minimum_last_seen_time_for_other_worker = value

    @property
    def worker_loop_sleep_time(self) -> float:
        with self._lock:
            return self._worker_loop_sleep_time

    @worker_loop_sleep_time.setter
    def worker_loop_sleep_time(self, value: float) -> None:
        with self._lock:
            self._worker_loop_sleep_time = value

    @property
    def worker_count(self) -> int:
        with self._lock:
            return len(self._workers)

    @property
    def worker_names(self) -> list[str]:
        with self._lock:
            return list(self._workers.keys())

    @property
    def workers(self) -> list[Worker]:
        with self._lock:
            return [Worker(object_id=worker_id) for worker_id in self._workers.values()]

    def create_worker(self,
                      name: _Optional[str] = None,
                      log_level: _Optional[int] = None,
                      log_push_rate: _Optional[float] = None,
                      log_push_max_entries: _Optional[int] = None,
                      loop_sleep_time: _Optional[float] = None,
                      task_ignore_loggers_equal: _Optional[list[str]] = None,
                      task_ignore_loggers_like: _Optional[list[str]] = None) -> Worker:
        self._logger.info(f"Creating worker '{name}'")

        worker_count = _worker_counter()

        if name is None:
            name = f"worker-{worker_count}"
        base_name = f"{WORKER_NAMESPACE_NAME}.{self.name}"
        if not name.startswith(base_name):
            name = f"{base_name}.{name}"

        if name in self.worker_names:
            raise ValueError(f"Worker with name '{name}' already exists")

        if log_level is None:
            log_level = self._logger.level

        # get worker document or create it
        try:
            worker_document = _WorkerDocument.objects.get(name=name,
                                                          manager=self.name)
            # check if other worker is running on this document
            if worker_document.state != WorkerState.QUIT:
                last_seen = _to_local(worker_document.last_seen)
                last_seen_delta_for_other_worker = _local_now() - last_seen
                if last_seen_delta_for_other_worker < _timedelta(seconds=self.minimum_last_seen_time_for_other_worker):
                    raise ValueError(f"Worker '{self.name}' already exists and was last seen {last_seen_delta_for_other_worker.total_seconds()} ago")
            worker_document.state = WorkerState.QUIT
            worker_document.started_at = None
            worker_document.last_seen = None
            worker_document.delay = None
            worker_document.current_task = None
            worker_document.signals = []
        except _DoesNotExist:
            worker_document = _WorkerDocument(name=name,
                                              manager=self.name,
                                              state=WorkerState.QUIT)
        worker_document.save()

        # create worker
        _WorkerThread(name=name,
                      manager=self,
                      document=worker_document,
                      log_level=log_level,
                      log_push_rate=log_push_rate,
                      log_push_max_entries=log_push_max_entries,
                      loop_sleep_time=loop_sleep_time,
                      task_ignore_loggers_equal=task_ignore_loggers_equal,
                      task_ignore_loggers_like=task_ignore_loggers_like)

        with self._lock:
            # add worker name and id to workers
            self._workers[name] = worker_document.id

        return Worker(object_id=worker_document.id)

    # --- task management ---

    @property
    def registered_tasks(self) -> list[str]:
        with self._lock:
            return list(self._registered_tasks.keys())

    @property
    def task_ignore_loggers_equal(self) -> list[str]:
        with self._lock:
            return self._task_ignore_loggers_equal.copy()

    @task_ignore_loggers_equal.setter
    def task_ignore_loggers_equal(self, value: list[str]):
        with self._lock:
            self._task_ignore_loggers_equal = value

    @property
    def task_ignore_loggers_like(self) -> list[str]:
        with self._lock:
            return self._task_ignore_loggers_like.copy()

    @task_ignore_loggers_like.setter
    def task_ignore_loggers_like(self, value: list[str]):
        with self._lock:
            self._task_ignore_loggers_like = value

    def get_task_func(self, name: str) -> callable:
        with self._lock:
            if name not in self._registered_tasks:
                raise ValueError(f"No task function with name '{name}' found")
            return self._registered_tasks[name]

    def registering_task(self, name: _Optional[str] = None, func: callable = None) -> callable:
        if self.worker_count > 0:
            raise RuntimeError("Cannot register tasks after workers are created")
        if not callable(func):
            raise ValueError(f"Argument 'func' must be a callable not '{type(func)}'")
        with self._lock:
            # get name from function if not provided
            if name is None:
                name = func.__name__

            self._logger.info(f"Registering task '{name}'")

            # check if task with name already registered
            if name in self._registered_tasks:
                raise ValueError(f"Task with name '{name}' already registered")

            # register task
            self._registered_tasks[name] = func
            self._registered_tasks_param[name] = {}
            self._registered_tasks_param_defaults[name] = {}

            # get signature from function
            func_params = dict(_inspect.signature(func).parameters)
            for param_name, param in func_params.items():
                self._registered_tasks_param[name][param_name] = param.annotation
                if param.default != _inspect.Parameter.empty:
                    self._registered_tasks_param_defaults[name][param_name] = param.default

            return func

    def register_task(self, name: _Optional[str] = None):
        def decorator(func: callable = None):
            return self.registering_task(name=name, func=func)

        return decorator

    def schedule_task(self, name: str, due: _Optional[_datetime] = None, log_level: _Optional[int] = None, **given_task_params) -> ScheduledTask:
        self._logger.info(f"Scheduling task '{name}'")

        if log_level is None:
            log_level = self._logger.level

        _ = self.get_task_func(name)
        if name not in self._registered_tasks_param:
            raise ValueError(f"No task parameters with name '{name}' found")
        task_params = self._registered_tasks_param[name]
        if name not in self._registered_tasks_param_defaults:
            raise ValueError(f"No task parameter defaults with name '{name}' found")
        task_param_defaults = self._registered_tasks_param_defaults[name]

        # check if all required parameters are provided
        for param_name, param_type in task_params.items():
            if param_name not in given_task_params:
                if param_name not in task_param_defaults:
                    raise ValueError(f"Parameter '{param_name}' is required for task '{name}'")
                try:
                    given_task_params[param_name] = task_param_defaults[param_name].copy()
                except AttributeError:
                    given_task_params[param_name] = task_param_defaults[param_name]

        # check if all provided parameters are correct
        for param_name, param_value in given_task_params.items():
            if not isinstance(param_value, task_params[param_name]):
                raise ValueError(f"Parameter '{param_name}' for task '{name}' must be of type '{task_params[param_name]}'")

        # create new task
        task = _TaskDocument(
            name=name,
            manager=self.name,
            state=TaskState.NEW,
            log_level=log_level,
            created_at=_local_now(),
            due_at=due or _local_now(),
            params=given_task_params
        )
        task.save()

        return ScheduledTask(object_id=task.id)



from wiederverwendbar.starlette_admin.mongoengine.boolean_also_field.converter import BooleanAlsoConverter
from wiederverwendbar.starlette_admin.mongoengine.generic_embedded_document_field.converter import GenericEmbeddedConverter
from wiederverwendbar.starlette_admin.mongoengine.ipv4_field.converter import IPv4Converter


class MongoengineConverter(GenericEmbeddedConverter, IPv4Converter, BooleanAlsoConverter):
    ...



from pydantic import Field

from wiederverwendbar.starlette_admin.settings.settings import AdminSettings


class FormMaxFieldsAdminSettings(AdminSettings):
    admin_form_max_fields: int = Field(default=1000, title="Form Max Fields", description="The maximum number of fields in a form.")



from wiederverwendbar.mongoengine.fields.boolean_also_field import BooleanAlsoField
from wiederverwendbar.mongoengine.fields.domain_field import DomainField
from wiederverwendbar.mongoengine.fields.ipv4_address_field import IPv4AddressField
from wiederverwendbar.mongoengine.fields.ipv4_network_field import IPv4NetworkField
from wiederverwendbar.mongoengine.fields.port_field import PortField
from wiederverwendbar.mongoengine.fields.with_instance_field import WithInstanceField

